# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import hashlib

import pydantic
from aiopki.types import Base64
from aiopki.utils import b64encode

from oauthx.lib.protocols import IClient
from oauthx.server.protocols import IResourceOwner
from oauthx.server.types import AuthorizationKey
from oauthx.server.types import TokenMAC


class AuthorizationCode(pydantic.BaseModel):
    aut: int
    client_id: str
    mac: Base64
    redirect_uri: str | None = None
    sub: int | str

    @property
    def authorization(self) -> AuthorizationKey:
        return AuthorizationKey(self.aut)

    @property
    def owner(self) -> IResourceOwner.Key:
        return IResourceOwner.Key(self.client_id, str(self.sub))

    def allows_redirect(self, redirect_uri: str | None) -> bool:
        if redirect_uri is not None:
            redirect_uri = b64encode(
                    hashlib.sha3_256(str.encode(redirect_uri or '')).digest(),
                    encoder=bytes.decode
            )
        return any([
            self.redirect_uri is None,
            str(self.redirect_uri) == redirect_uri
        ])

    def is_authorized(self, client_id: str) -> bool:
        return self.client_id == client_id

    def is_revoked(self, client: IClient, owner: IResourceOwner) -> bool:
        mac = TokenMAC()
        mac.add(client)
        mac.add(owner)
        return bytes(mac) == self.mac


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .basestorage import BaseStorage
from .plugin import Plugin
from .oidcrouter import OIDCRouter
from .params import Storage
from .policy import Policy
from .subjectlogger import SubjectLogger


__all__: list[str] = [
    'create_server',
    'BaseStorage',
    'Plugin',
    'OIDCRouter',
    'Policy',
    'Storage',
    'SubjectLogger',
]

create_server = OIDCRouter.create_server


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from canonical import UnixTimestamp

from .redirectionparameters import RedirectionParameters


class JARMAuthorizationResponse(RedirectionParameters):
    aud: str
    iss: str # type: ignore
    exp: UnixTimestamp
    jti: str | None
    iat: UnixTimestamp | None = None
    nbf: UnixTimestamp | None = None




# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
from typing import Any

import pydantic

from oauthx.lib.types import AccessTokenHash


class TokenSubject(pydantic.BaseModel):
    """Maintains details about a :term:`Subject` as retrieved from the UserInfo
    endpoint of an authorization server.
    """
    at_hash: AccessTokenHash = pydantic.Field(
        default=...,
        title="Hash",
        description=(
            "The SHA3-256 hash of the access token that was used to "
            "authenticate the subject."
        )
    )
    
    exp: datetime.datetime = pydantic.Field(
        default_factory=lambda: (
            datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(seconds=3600)
        ),
        title="Expires",
        description=(
            "Indicates the date and time at which the maintained "
            "user information expires."
        )
    )

    sub: str = pydantic.Field(
        default=...,
        title="Subject ID",
        description=(
            "The subject identifier, as provided by the authorization "
            "server."
        )
    )
    
    claims: dict[str, Any] = pydantic.Field(
        default_factory=dict,
        title="Claims",
        description=(
            "The claims that the access token could obtain for the "
            "subject."
        )
    )
    
    @property
    def pk(self) -> str:
        return self.at_hash


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from oauthx.lib.exceptions import Error as ServerError


class Error(pydantic.BaseModel):
    error: str
    error_description: str | None = None
    error_uri: str | None = None
    state: str | None = None
    iss: str | None = None

    def is_error(self) -> bool:
        return True

    def raise_for_status(self):
        raise ServerError(**self.model_dump())


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Union

import pydantic

from .authorizationrequestobject import AuthorizationRequestObject
from .authorizationrequestparameters import AuthorizationRequestParameters
from .authorizationrequestreference import AuthorizationRequestReference


AuthorizationRequestType = Union[
    AuthorizationRequestParameters,
    AuthorizationRequestObject,
    AuthorizationRequestReference
]


class AuthorizationRequest(pydantic.RootModel[AuthorizationRequestType]):
    pass


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import TypeVar

from canonical import ResourceIdentifier

from oauthx.lib.types import AccessTokenHash
from oauthx.types import IStorage
from oauthx.types import Masked
from oauthx.resource import TokenSubject
from oauthx.resource import IResourceServerStorage
from .datastorestorage import DatastoreStorage


T = TypeVar('T')


class ResourceServerStorage(DatastoreStorage, IStorage, IResourceServerStorage):

    async def get(self, key: ResourceIdentifier[Any, T]) -> T | None:
        raise NotImplementedError

    async def get_token_subject(self, at_hash: AccessTokenHash) -> TokenSubject | None:
        return await self.get_model_by_key(TokenSubject, str(at_hash))

    async def mask(self, obj: Any) -> Masked:
        raise NotImplementedError

    async def persist(self, object: Any) -> None:
        await self.put(object, object.pk)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from oauthx.lib.types import RedirectURI


class InvalidAuthorizationRequest(Exception):
    __module__: str = 'oauthx.types'
    allow_redirect: bool = True
    kind: str = 'authorize'
    media_type: str | None = None
    redirect_uri: RedirectURI | None = None
    status_code: int = 400
    template_name: str = 'oauthx/errors/authorize.html.j2'

    def __init__(
        self,
        *,
        error: str,
        context: Any,
        error_description: str | None = None,
        error_uri: str | None = None,
        redirect_uri: RedirectURI | None = None,
        allow_redirect: bool = True
    ):
        self.allow_redirect = allow_redirect
        self.context = context
        self.error = error
        self.error_description = error_description
        self.error_uri = error_uri
        self.redirect_uri = redirect_uri

    def get_context(self) -> dict[str, Any]:
        return {}

    def get_template_names(self) -> list[str] | str:
        return [
            f'oauthx/errors/{self.kind}.{self.error}.html.j2',
            self.template_name
        ]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import json
import os
import pathlib
import urllib.parse
from typing import Any
from typing import Awaitable
from typing import Callable

import fastapi
import httpx
import pytest
import pytest_asyncio
import aiopki
import yaml
from aiopki.lib import JSONWebKey
from aiopki.utils import b64encode

from oauthx.client import Client as ConsumingClient
from oauthx.lib import AuthorizationResponse
from oauthx.lib import ClientAuthorizationState
from oauthx.lib import Error
from oauthx.lib import RedirectionParameters
from oauthx.server.config import Config
from oauthx.server.params.masker import _Masker as Masker # type: ignore
from oauthx.server import OIDCRouter
from oauthx.server.ref import Storage
from oauthx.server.models import ObjectFactory
from oauthx.server.models import ResourceOwner
from oauthx.server.models import Session
from oauthx.server.request import Request
from oauthx.server.models import Client as ServerClient


aiopki.install_backend('aiopki.ext.cryptography')


@pytest_asyncio.fixture(scope='function') # type: ignore
async def agent(app: fastapi.FastAPI, session_cookie: str):
    params = {
        'cookies': {'sessionid':  session_cookie}
    }
    async with httpx.AsyncClient(app=app, **params) as client: # type: ignore
        yield client


@pytest_asyncio.fixture # type: ignore
async def app(
    config: Config,
    storage_class: type[Storage]
):
    app = fastapi.FastAPI()
    client = httpx.AsyncClient(app=app)
    await client.__aenter__()
    router = OIDCRouter.from_config(
        app=app,
        config=config,
        prefix='/oauth/v2',
        http=client
    )
    app.include_router(router)
    yield app
    await client.__aexit__(None, None, None)



@pytest_asyncio.fixture(scope='function') # type: ignore
async def authorize_request(
    client: ConsumingClient,
    agent: httpx.AsyncClient,
) -> Callable[..., Awaitable[tuple[AuthorizationResponse | None, Error| RedirectionParameters | None, ClientAuthorizationState | None]]]:
    #await client.discover(agent)

    async def f(
        *,
        status_code: int = 302,
        expect_error: str | None = None,
        expect_params: set[str] = set(),
        expect_media_type: str | None = None,
        **kwargs: Any
    ) -> tuple[AuthorizationResponse | None, Error| RedirectionParameters  | None, ClientAuthorizationState | None]:
        state = await client.authorize(
            http=agent,
            state_class=ClientAuthorizationState,
            **kwargs
        )
        response = await agent.get(state.authorize_url)
        assert response.status_code == status_code,\
            f"{response.headers.get('X-Error')}: {response.headers.get('X-Error-Description')}\n"\
            + '\n'.join([f'{k}: {v}' for k, v in sorted(response.headers.items(), key=lambda x: x[0])])
        if 300 <= status_code < 400:
            assert 'Location' in response.headers
            assert 'X-Error' not in response.headers or expect_error,\
                f"{response.headers.get('X-Error')}: {response.headers.get('X-Error-Description')}"
            assert 'X-Error' in response.headers or not expect_error,\
                f"{response.headers.get('X-Error')}: {response.headers.get('X-Error-Description')}"
            assert not expect_media_type or response.headers.get('Content-Type') == expect_media_type
            try:
                loc: str = response.headers['Location']
                p = urllib.parse.urlparse(loc)
                q = dict(urllib.parse.parse_qsl(p.query))
                assert set(q.keys()) == expect_params or not expect_params,\
                    f'Missing parameters: {str.join(",", expect_params - set(q.keys()))}'
                return AuthorizationResponse.model_validate(q), await client.on_redirected(state, q), state
            except json.JSONDecodeError:
                for name, value in response.headers.items():
                    print(f'{name}: {value}')
                print(response.content)
                raise

        return None, None, None

    return f


@pytest_asyncio.fixture(scope='function') # type: ignore
async def client(
    client_credential: Any,
    response_mode: str,
    client_params: dict[str, Any]
) -> ConsumingClient:
    return ConsumingClient.model_validate({
        'provider': 'http://127.0.0.1',
        'client_id': 'foo',
        'credential': client_credential,
        'response_mode': response_mode,
        **client_params
    })


@pytest_asyncio.fixture(scope='function') # type: ignore
async def client_credential() -> Any:
    return None


@pytest_asyncio.fixture(scope='function') # type: ignore
async def client_params() -> dict[str, Any]:
    return {}


@pytest.fixture(scope='function')
def config(config_data: dict[str, Any]) -> Config:
    return Config.model_validate(config_data)


@pytest.fixture(scope='session')
def config_data() -> dict[str, Any]:
    data = yaml.safe_load(open(pathlib.Path(__file__).parent.parent.joinpath('oidc.conf'), 'r').read())
    data['issuer']['id'] = 'http://127.0.0.1'
    return data


@pytest_asyncio.fixture(scope='function') # type: ignore
async def dek(config: Config) -> aiopki.CryptoKeyType:
    return await config.storage.encryption_key


@pytest.fixture(scope='function')
def grant_types() -> list[str]:
    return ['authorization_code']


@pytest_asyncio.fixture(scope='function') # type: ignore
async def http(app: fastapi.FastAPI):
    async with httpx.AsyncClient(app=app) as client: # type: ignore
        yield client


@pytest.fixture
def redirect_uris() -> list[str]:
    return ['http://127.0.0.1:23390']


@pytest.fixture(scope='function')
def response_mode() -> str:
    return 'query'


@pytest.fixture(scope='function')
def response_types() -> list[str]:
    return ['code']


@pytest.fixture(scope='function')
def scope() -> list[str]:
    return ['email']


@pytest_asyncio.fixture # type: ignore
async def session_signer(config: Config) -> Any:
    return await config.ui.session_key # type: ignore


@pytest_asyncio.fixture # type: ignore
async def session_cookie(session_signer: aiopki.CryptoKeyType) -> str:
    session = Session.new()
    session.sub = '1'
    return await Request.sign_session(session, session_signer) # type: ignore


@pytest_asyncio.fixture(scope='function') # type: ignore
async def signer(config: Config) -> aiopki.CryptoKeyType:
    await config.issuer.signing_key
    return config.issuer.signing_key


@pytest_asyncio.fixture(autouse=True) # type: ignore
async def storage(
    config: Config,
    storage_class: type[Storage],
    dek: JSONWebKey
) -> Storage:
    storage = storage_class(config=config, key=dek)
    factory = ObjectFactory(
        issuer=config.issuer.id,
        storage=storage,
        masker=Masker(
            key=JSONWebKey.model_validate({ # type: ignore
                'kty': 'oct',
                'use': 'sig',
                'alg': 'HS384',
                'k': b64encode(os.urandom(32))
            })
        )
    )
    await storage.persist(
        ResourceOwner.model_validate({
            'client_id': 'default',
            'sub': '1',
            'sector_identifier': 'localhost'
        }),
    )
    subject = await factory.subject(pk=1, use='personal')
    await subject.encrypt_keys(dek)
    await storage.persist(subject)
    return storage


@pytest.fixture(scope='function')
def storage_class() -> type[Storage]:
    return type('Storage', (Storage, ), {'objects': {}})


@pytest_asyncio.fixture(scope='function', autouse=True) # type: ignore
async def server_client(
    client_credential: Any,
    http: httpx.AsyncClient,
    storage: Storage,
    grant_types: list[str],
    redirect_uris: list[str],
    response_types: list[str],
    response_mode: str,
    scope: list[str],
    client_params: dict[str, Any]
) -> ServerClient:
    client = ServerClient.model_validate({
        'client_id': 'foo',
        'client_name': 'foo',
        'grant_types': grant_types,
        'scope': scope,
        'redirect_uris': redirect_uris,
        'response_modes': [response_mode],
        'response_types': response_types,
        'credential': client_credential,
        'access_tokens': {
            'claims': ['email'],
            'ttl': 600
        },
        **client_params
    })
    await client.persist(storage)
    return client


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import secrets
from typing import Any

import httpx
import pytest
import pytest_asyncio

from aiopki.ext.jose import JWKS
from oauthx.models import Client


@pytest.fixture(scope='function')
def grant_types() -> list[str]:
    return ['urn:ietf:params:oauth:grant-type:jwt-bearer']


@pytest.fixture(scope='function')
def response_types() -> list[str]:
    return ['none']


@pytest_asyncio.fixture(scope='function') # type: ignore
async def client_credential() -> Any:
    return JWKS.model_validate({
        'keys': [{
            'kty': 'oct',
            'use': 'sig',
            'alg': 'HS256',
            'k': secrets.token_urlsafe(32)
        }]
    })


@pytest.mark.asyncio
async def test_jwt_bearer(
    agent: httpx.AsyncClient,
    http: httpx.AsyncClient,
    client: Client
):
    # Authorize the client
    state = await client.authorize(response_type='none')
    response = await agent.get(state.authorize_url, follow_redirects=False)
    assert response.status_code == 302, response.content
    await client.jwt('1', http=http)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi

from oauthx.lib.exceptions import InvalidToken
from oauthx.resource import AccessToken
from oauthx.server.models import SubjectKey
from oauthx.server.params import Storage
from oauthx.server.protocols import ISubject


async def get(at: AccessToken, storage: Storage) -> ISubject:
    subject = await storage.get(SubjectKey(at.sub))
    if subject is None:
        raise InvalidToken(
            "The subject identified by the access token does not "
            "exist."
        )
    return subject


Subject: TypeAlias = Annotated[ISubject, fastapi.Depends(get)]




# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import MutableMapping
from typing import Protocol

from aiopki.types import EncryptionResult
from aiopki.types import IDecrypter
from aiopki.types import IEncrypter
from aiopki.types import ISigner
from aiopki.types import Plaintext
from aiopki.ext.jose import OIDCToken
from canonical import ResourceIdentifier


class ISubject(Protocol):
    """Specifies the interface of a model that represents a
    :term:`Subject`
    """
    __module__: str = 'oauthx.types'

    def is_encrypted(self) -> bool:
        ...

    async def decrypt_keys(self, key: IDecrypter) -> None:
        """Decrypt the subjects' encryption keys using the
        storage encryption key.
        """
        ...

    async def encrypt_keys(self, key: IEncrypter) -> None:
        """Encrypt the subjects' encryption keys using the
        storage encryption key.
        """
        ...

    async def decrypt(self, value: EncryptionResult) -> Plaintext:
        ...

    async def encrypt(self, value: bytes) -> EncryptionResult:
        ...

    async def mask(self, value: bytes) -> str:
        ...

    def contribute_to_userinfo(self, userinfo: MutableMapping[str, Any]) -> None:
        ...

    def get_primary_key(self) -> ResourceIdentifier[int | str, 'ISubject']:
        """Return the primary key of the :class:`ISubject`."""
        ...

    def get_masking_key(self) -> ISigner:
        ...

    def update_from_oidc(self, token: OIDCToken) -> None:
        ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Union
from typing import TypeAlias

import pydantic

from .jwtbearerassertion import JWTBearerAssertion
from .saml2bearerassertion import SAML2BearerAssertion


__all__: list[str] = ['Assertion']


AssertionType: TypeAlias = Union[
    JWTBearerAssertion,
    SAML2BearerAssertion
]


class Assertion(pydantic.RootModel[AssertionType]):
    pass


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import os
from typing import Iterable

from oauthx.client.models import Client
from oauthx.lib.params import HTTPClient
from .clientresolver import ClientResolver
from .models import Client


class DefaultClient(ClientResolver):
    __module__: str = 'oauthx.client'
    client: Client | None = None

    def __init__(
        self,
        scope: Iterable[str] | None = None
    ):
        self.scope = scope

    async def get(self, client_id: str) -> Client:
        assert self.client is not None
        return self.client

    async def resolve(self, http: HTTPClient) -> Client: # type: ignore
        if self.client is None:
            self.client = Client.model_validate({
                'provider': os.getenv('OAUTH2_ISSUER'),
                'client_id': os.getenv('OAUTH2_CLIENT_ID'),
                'scope': self.scope
            })
            await self.client.discover(http=http)
        return self.client


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from .clientauthorizationstatekey import ClientAuthorizationStateKey


class RedirectionParameters(pydantic.BaseModel):
    code: str
    state: ClientAuthorizationStateKey | None = None
    iss: str | None = None

    @property
    def error(self) -> str:
        raise NotImplementedError

    @property
    def error_description(self) -> str:
        raise NotImplementedError

    @property
    def error_uri(self) -> str:
        raise NotImplementedError

    def is_error(self) -> bool:
        return False

    def raise_for_status(self):
        pass


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .fatalerror import FatalError


class InsufficientScope(FatalError, ValueError):
    __module__: str = 'oauthx.types'
    status_code: int = 403
    default_description: str = (
        "The request requires higher privileges than "
        "provided by the access token."
    )

    def __init__(self, description: str | None = None, url: str | None = None, allow_redirect: bool = False):
        FatalError.__init__(
            self,
            error='insufficient_scope',
            error_description=description or self.default_description,
            error_uri=url,
            allow_redirect=allow_redirect
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal

from aiopki.types import SubjectID
from aiopki.types import EncryptionResult

from .baseprincipal import BasePrincipal


class SubjectIdentifierPrincipal(BasePrincipal):
    kind: Literal['subject'] = 'subject'
    value: EncryptionResult | SubjectID
    
    def to_bytes(self) -> bytes:
        assert not isinstance(self.value, EncryptionResult)
        return str.encode(self.value.model_dump_json())


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
import secrets
from typing import Any
from typing import ClassVar
from typing import Callable

import pydantic
from aiopki.types import Base64

from oauthx.server.protocols import IResourceOwner
from .subjectkey import SubjectKey


class ResourceOwner(pydantic.BaseModel):
    Key: ClassVar = IResourceOwner.Key

    blocked: bool = pydantic.Field(
        default=False
    )

    client_id: str  = pydantic.Field(
        default=...
    )

    key: Base64 = pydantic.Field(
        default_factory=lambda: secrets.token_bytes(32)
    )

    sector_identifier: str = pydantic.Field(
        default=...
    )
    
    scope: set[str] = pydantic.Field(
        default_factory=set
    )

    sub: SubjectKey = pydantic.Field(
        default=...
    )

    onboarded: datetime.datetime = pydantic.Field(
        default_factory=lambda: datetime.datetime.now(datetime.timezone.utc)
    )

    @property
    def pk(self) -> Key:
        return self.Key(self.client_id, str(self.sub))

    def grants_consent(self, scope: set[str]) -> bool:
        return True

    def model_dump_json(self, **kwargs: Any) -> str:
        exclude = kwargs.setdefault('exclude', set())
        exclude.update({
            'claims',
            'principals'
        })
        return super().model_dump_json(**kwargs)

    def update_key(self, update: Callable[[bytes], None]):
        update(self.key)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi

from oauthx.lib.params import HTTPClient
from oauthx.lib.responses import UserInfoResponse
from .bearertoken import BearerToken
from .provider import Provider


__all__: list[str] = ['UserInfo']


async def get(provider: Provider, token: BearerToken, http: HTTPClient) -> UserInfoResponse:
    return await provider.userinfo(token, http)


UserInfo: TypeAlias  = Annotated[UserInfoResponse, fastapi.Depends(get)]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi

from oauthx.lib.params import Logger
from oauthx.server.types import AuthorizationRequestKey
from oauthx.server.types import IAuthorizationRequest
from oauthx.server.types import StopSnooping
from .storage import Storage


__all__: list[str] = ['PendingAuthorizationRequest']


async def get(
    logger: Logger,
    storage: Storage,
    pk: AuthorizationRequestKey = fastapi.Path(...)
) -> IAuthorizationRequest:
    request = await storage.get(pk)
    if request is None:
        logger.debug("Authorization request does not exist (id: %s)", int(pk))
        raise StopSnooping
    return request
    

PendingAuthorizationRequest: TypeAlias =  Annotated[IAuthorizationRequest, fastapi.Depends(get)]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from canonical import ResourceIdentifier

from .iauthorization import IAuthorization


class AuthorizationKey(ResourceIdentifier[int, IAuthorization]):
    __module__: str = 'oauthx.server.types'
    
    def __init__(self, pk: int):
        self.__pk = pk

    def __int__(self) -> int:
        return int(self.__pk)

    def __str__(self) -> str:
        return str(self.__pk)

    def __hash__(self) -> int:
        return hash(self.__pk)

    def __repr__(self) -> str:
        return f'<AuthorizationKey: {str(self)}>'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi

from oauthx.server import models
from oauthx.server.params import CurrentSubject
from oauthx.server.params import Storage
from oauthx.server.protocols import IResourceOwner
from .client import Client
from .redirecturi import RedirectURI


__all__: list[str] = [
    'ResourceOwner'
]


async def get(
    storage: Storage,
    client: Client,
    subject: CurrentSubject,
    redirect_uri: RedirectURI
) -> IResourceOwner | None:
    if subject is None:
        return None
    owner = models.ResourceOwner.model_validate({
        'client_id': str(client.id), # type: ignore
        'sub': subject.get_primary_key(),
        'sector_identifier': client.get_sector_identifier()
    })
    await storage.persist(owner)
    return owner


ResourceOwner: TypeAlias = Annotated[
    IResourceOwner | None,
    fastapi.Depends(get)
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi
from aiopki import CryptoKeyType

from oauthx.server.config import Config
from .currentconfig import CurrentConfig


__all__: list[str] = ['ContentEncryptionKey']


async def get(config: Config = CurrentConfig) -> CryptoKeyType:
    cek = config.get_cek()
    await cek
    return cek


ContentEncryptionKey: TypeAlias = Annotated[CryptoKeyType, fastapi.Depends(get)]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

from oauthx.resource import ResourceRoute
from oauthx.server.config import Config
from oauthx.server.request import Request


class UserInfoRoute(ResourceRoute):
    config: Config

    def setup_request(self, request: fastapi.Request) -> Request:
        request = Request(request.scope, request.receive)
        return request


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Protocol

from oauthx.lib.protocols import IStorage


class IAuthorization(Protocol):
    __module__: str = 'oauthx.server.types'

    @property
    def pk(self) -> Any:
        ...

    async def consume(self, storage: IStorage) -> None:
        """Consume the :class:`IAuthorization` and mark its
        authorization code as used.
        """
        ...

    def is_consumed(self) -> bool:
        """Return a boolean indicating if the :class:`IAuthorization`
        is consumed i.e. it's authorization code was exchanged.
        """
        ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal

import pydantic
from aiopki.ext.jose import CompactSerialized

from oauthx.lib.types import ScopeType
from .basegrant import BaseGrant


class JWTBearerGrant(BaseGrant):
    grant_type: Literal['urn:ietf:params:oauth:grant-type:jwt-bearer'] = pydantic.Field(
        default=...,
        description=(
            "This value **must** be `urn:ietf:params:"
            "oauth:grant-type:jwt-bearer`."
        )
    )
    
    assertion: CompactSerialized = pydantic.Field(
        default=...,
        description=(
            "A compact-encoded JSON Web Token (JWT) that is signed, "
            "containing the claims specified in RFC 7523."
        )
    )

    scope: ScopeType = pydantic.Field(
        default_factory=ScopeType,
        description=(
            "The space-delimited scope of the access request."
        )
    )
    
    def must_authenticate(self) -> bool:
        return False

    def must_identify(self) -> bool:
        return False


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .accesstokenhash import AccessTokenHash
from .clientassertiontype import ClientAssertionType
from .clientcredentialtype import ClientCredentialType
from .masked import Masked
from .oidcissueridentifier import OIDCIssuerIdentifier
from .granttype import GrantType
from .prompttype import PrompType
from .redirecturi import RedirectURI
from .requesturi import RequestURI
from .responsemodetype import ResponseModeType
from .responsetype import ResponseType
from .scopetype import ScopeType
from .subjectclaimtype import SubjectClaimType
from .accesstoken import AccessToken
from .clientassertiontype import ClientAssertionType
from .clientauthenticationmethod import ClientAuthenticationMethod
from .clientsecret import ClientSecret
from .iuserinterface import IUserInterface
from .masked import Masked
from .maskedprincipal import MaskedPrincipal
from .nullredirecturi import NullRedirectURI
from .pckechallengemethod import PKCEChallengeMethod
from .sectoridentifieruri import SectorIdentifierURI
from .tokentype import TokenType
from .unmasked import Unmasked


__all__: list[str] = [
    'AccessTokenHash',
    'ClientAssertionType',
    'ClientCredentialType',
    'OIDCIssuerIdentifier',
    'GrantType',
    'Masked',
    'PrompType',
    'RedirectURI',
    'RequestURI',
    'ResponseModeType',
    'ResponseType',
    'ScopeType',
    'SubjectClaimType',
    'AccessToken',
    'ClientAuthenticationMethod',
    'ClientAssertionType',
    'ClientSecret',
    'IUserInterface',
    'Masked',
    'MaskedPrincipal',
    'NullRedirectURI',
    'PKCEChallengeMethod',
    'SectorIdentifierURI',
    'TokenType',
    'Unmasked',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from canonical import ResourceIdentifier

from oauthx.lib.types import OIDCIssuerIdentifier
from .iclaimset import IClaimSet


class ClaimSetKey(ResourceIdentifier[int, IClaimSet]):
    __module__: str = 'oauthx.server.types'
    
    def __init__(self, iss: OIDCIssuerIdentifier, sub: int):
        self.iss = iss
        self.sub = sub

    def __hash__(self) -> int:
        return hash((self.iss, self.sub))

    def __repr__(self) -> str:
        return f'<ClaimSetKey: {self.iss}/sub/{self.sub}>'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
from typing import Literal
from typing import Protocol

from oauthx.lib.types import OIDCIssuerIdentifier


class ISubjectLogger(Protocol):
    __module__: str = 'oauthx.server.types'

    def onboarded(
        self,
        sub: int,
        registrar: str,
        authorization_id: int | None = None,
        timestamp: datetime.datetime | None = None,
        client_id: str | None = None,
    ) -> None:
        ...

    def receipt(
        self,
        receipt_id: int,
        sub: int,
        obtained: datetime.datetime,
        processed: set[str],
        provider: OIDCIssuerIdentifier,
        purpose: Literal['IDENTIFY', 'INVITE', 'LOGIN', 'VERIFY_ACCOUNT'],
        received: set[str],
        client_id: str | None = None,
        request_id: int | None = None,
        timestamp: datetime.datetime | None = None
    ) -> None:
        ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import urllib.parse
from gettext import gettext as _
from typing import Any

import fastapi

from oauthx.server.config import Config
from oauthx.server.config import ProviderConfig
from oauthx.server.params import AuthorizationServer
from oauthx.server.params import CurrentConfig
from oauthx.server.params import CurrentSubject
from oauthx.server.params import RequestSession
from oauthx.server.params import Storage
from oauthx.server.request import Request
from oauthx.server.types import StopSnooping
from ..baserequesthandler import BaseRequestHandler
from .params import AuthorizationRequest
from .params import Client


class BaseLoginRequestHandler(BaseRequestHandler):
    __module__: str = 'oauthx.server.handlers'
    client: Client
    name: str = 'oauth2.login'
    next_url: str | None = None
    path: str = '/login'
    return_url: str | None = None
    response_class: type[fastapi.Response] = fastapi.responses.RedirectResponse
    response_description: str = "Proceed with authentication"
    session: RequestSession
    subject: CurrentSubject
    summary: str = "Login Endpoint"

    def __init__(
        self,
        server: AuthorizationServer,
        request: Request,
        client: Client,
        params: AuthorizationRequest,
        session: RequestSession,
        storage: Storage,
        subject: CurrentSubject,
        config: Config = CurrentConfig,
        deny_url: str | None = fastapi.Query(
            default=None,
            alias='d',
            title="Deny URL"
        ),
        next_url: str | None = fastapi.Query(
            default=None,
            alias='n',
            title="Next URL"
        ),
        return_url: str | None = fastapi.Query(
            default=None,
            alias='b',
            title="Return URL"
        ),
        **kwargs: Any
    ):
        self.client = client
        self.deny_url = deny_url
        self.next_url = next_url
        self.params = params
        self.return_url = return_url
        self.session = session
        self.subject = subject
        super().__init__(server=server, request=request, storage=storage, config=config, **kwargs)

    def allows_next_url(self, request: Request, url: str | None) -> bool:
        try:
            p = urllib.parse.urlparse(url)
            return all([
                p.scheme == request.url.scheme,
                p.netloc == request.url.netloc
            ])
        except Exception:
            return False

    def get_context(self, request: fastapi.Request) -> dict[str, Any]:
        assert self.client is not None
        return {
            'begin_login_url': request.url_for('oauth2.upstream.begin'),
            'client_id': self.client.client_id,
            'client_name': self.client.get_display_name(),
            'deny_url': self.deny_url,
            'next_url': self.next_url,
            'page_title': _("Sign in"),
            'page_subtitle': _("to continue to **%s**") % self.client.get_display_name(),
            'params': self.params,
            'providers': self.get_providers(),
            'request': self.request
        }

    def get_providers(self) -> list[ProviderConfig]:
        return self.config.providers

    def get_templates(self) -> list[str]:
        return ['oauthx/login.upstream.html.j2']\
            if self.config.providers\
            else ['oauthx/login.html.j2']

    async def prepare(self, request: Request) -> fastapi.Response | None:
        if None in {self.next_url}:
            return await request.render_to_response('oauthx/404.html.j2', status_code=404)
        if any([
            not self.allows_next_url(request, self.next_url),
            self.client is None
        ]):
            raise StopSnooping

    async def handle(self, request: Request) -> fastapi.Response:
        assert self.client is not None
        assert self.next_url is not None
        return await request.render_to_response(
            self.get_templates(),
            context=self.get_context(self.request)
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import threading
import time
from typing import Any

import fastapi
import uvicorn

from oauthx.lib.models import ClientAuthorizationState
from oauthx.lib.models import ObtainedGrant
from oauthx.client import Client


class LocalRedirection(uvicorn.Server):
    grant: ObtainedGrant | None = None

    def __init__(
        self,
        client: Client,
        state: ClientAuthorizationState,
        port: int
    ):
        self.client = client
        self.state = state
        self.port = port
        self.app = fastapi.FastAPI()
        self.app.add_api_route('/', self.callback, methods=['GET'])
        self.config = uvicorn.Config(
            app=self.app,
            port=self.port,
            access_log=False,
            log_level='error',
        )
        self.event = threading.Event()
        super().__init__(self.config)

    async def callback(self, request: fastapi.Request) -> fastapi.Response:
        self.event.set()
        self.result = await self.client.on_redirected(
            state=self.state,
            params=dict(request.query_params)
        )
        if self.result.is_error():
            return fastapi.responses.PlainTextResponse("An error occurred.")
        self.grant = await self.client.authorization_code(self.result, self.state)
        return fastapi.responses.PlainTextResponse("Hello world!")

    def install_signal_handlers(self):
        pass

    def raise_for_status(self):
        self.result.raise_for_status()

    def __enter__(self):
        self.thread = threading.Thread(target=self.run)
        self.thread.start()
        while not self.started:
            time.sleep(1e-3)
        return self

    def __exit__(self, cls: type[Exception], *args: Any):
        self.event.wait()
        self.should_exit = True
        self.thread.join()
        if cls:
            raise NotImplementedError



# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import os

import aiopki
import jinja2
import pydantic
import yaml
from canonical import PythonSymbol

from oauthx.lib.protocols import IStorage
from oauthx.lib.types import GrantType
from oauthx.lib.utils.loader import import_symbol
from oauthx.server.models import Client
from oauthx.server.protocols import IRequestSubject
from oauthx.server.protocols import IPlugin
from .authenticationconfig import AuthenticationConfig
from .cacheconfig import CacheConfig
from .implementationconfig import ImplementationConfig
from .issuerconfig import IssuerConfig
from .providerconfig import ProviderConfig
from .scopespecification import ScopeSpecification
from .storageconfig import StorageConfig
from .uiconfig import UIConfig


class Config(pydantic.BaseModel):
    authentication: AuthenticationConfig
    cache: CacheConfig = pydantic.Field(
        default_factory=CacheConfig
    )

    client: Client | None
    grant_types_supported: set[GrantType] = set()
    impl: ImplementationConfig
    issuer: IssuerConfig
    plugins: list[PythonSymbol[type[IPlugin]]] = []
    providers: list[ProviderConfig] = []
    scopes: list[ScopeSpecification] = []
    storage: StorageConfig
    ui: UIConfig

    @property
    def storage_class(self) -> type[IStorage]:
        return import_symbol(self.impl.storage)

    @property
    def subject_class(self) -> type[IRequestSubject]:
        return import_symbol(self.impl.subject)

    @classmethod
    def load(cls, fp: str) -> 'Config':
        with open(fp, 'r') as f:
            t = jinja2.Template(
                source=f.read(),
                undefined=jinja2.StrictUndefined,
                variable_start_string='${',
                variable_end_string='}'
            )
        c = {
            'env': dict(os.environ)
        }
        return cls.model_validate(yaml.safe_load(t.render(c)))

    async def discover(self):
        await asyncio.gather(*map(ProviderConfig.discover, self.providers))
        return self

    def get_cek(self) -> aiopki.CryptoKeyType:
        return self.storage.encryption_key

    def get_splash_image_url(self) -> str:
        return 'https://static.webiam.id/shared/splash.jpg'

    def get_provider(self, name: str) -> ProviderConfig | None:
        return {x.name: x for x in self.providers}.get(name)

    def has_authorize_endpoint(self) -> bool:
        """Return a boolean indicating if the authorization server should
        expose the Authorization Endpoint.
        """
        return 'authorization_code' in self.grant_types_supported
    
    def __await__(self):
        return self.discover().__await__()



# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Declares :class:`ServerMetadata`."""
from typing import Any

import httpx
import pydantic
from aiopki.lib import JSONWebKeySet

from oauthx.lib import utils
from oauthx.lib.responses import UserInfoResponse
from oauthx.lib.types import ClientAuthenticationMethod


class ServerMetadata(pydantic.BaseModel):
    issuer: str = pydantic.Field(
        default=...,
        title="Issuer",
        description="Authorization server's issuer identifier URL.",
    )

    authorization_endpoint: str | None = pydantic.Field(
        default=None,
        title="Authorization endpoint",
        description="URL of the authorization server's authorization endpoint.",
    )

    token_endpoint: str = pydantic.Field(
        default=...,
        title="Token endpoint",
        description="URL of the authorization server's token endpoint.",
    )

    jwks_uri: str | None = pydantic.Field(
        default=None,
        title="JSON Web Key Set (JWKS) URI",
        description="URL of the authorization server's JWK Set document.",
    )

    registration_endpoint: str | None = pydantic.Field(
        default=None,
        title="Registration endpoint",
        description=(
            "URL of the authorization server's **OAuth 2.0 Dynamic Client "
            "Registration Endpoint**."
        ),
    )

    scopes_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported scopes",
        description=(
            "JSON array containing a list of the OAuth 2.0 `scope` values that "
            "this authorization server supports."
        ),
    )

    response_types_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported response types",
        description=(
            "JSON array containing a list of the OAuth 2.0 `response_type` "
            "values that this authorization server supports."
        ),
    )

    response_modes_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported response modes",
        description=(
            "JSON array containing a list of the OAuth 2.0 `response_mode` "
            "values that this authorization server supports."
        ),
    )

    grant_types_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported grant types",
        description=(
            "JSON array containing a list of the OAuth 2.0 `grant_types` "
            "values that this authorization server supports."
        ),
    )

    token_endpoint_auth_methods_supported: list[ClientAuthenticationMethod] = pydantic.Field(
        default=[],
        title="Supported client authentication methods",
        description=(
            "JSON array containing a list of client authentication methods "
            "supported by this token endpoint."
        ),
    )

    token_endpoint_auth_signing_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the token endpoint for the signature on the JWT "
            "used to authenticate the client at the token endpoint."
        ),
    )

    service_documentation: str | None = pydantic.Field(
        default=None,
        title="Documentation",
        description=(
            "URL of a page containing human-readable information that "
            "developers might want or need to know when using the "
            "authorization server."
        ),
    )

    ui_locales_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported locals for UI",
        description=(
            "Languages and scripts supported for the user interface, "
            "represented as a JSON array of language tag values from BCP 47."
        ),
    )

    op_policy_uri: str | None= pydantic.Field(
        default=None,
        title="Data policy URL",
        description=(
            "URL that the authorization server provides to the person "
            "registering the client to read about the authorization server's "
            "requirements on how the client can use the data provided by the "
            "authorization server."
        ),
    )

    op_tos_uri: str | None= pydantic.Field(
        default=None,
        title="Terms of service URL",
        description=(
            "URL that the authorization server provides to the person "
            "registering the client to read about the authorization server's "
            "terms of service."
        ),
    )

    revocation_endpoint: str | None = pydantic.Field(
        default=None,
        title="Revocation endpoint",
        description="URL of the authorization server's revocation endpoint.",
    )

    revocation_endpoint_auth_methods_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported authentication methods",
        description=(
            "JSON array containing a list of client authentication methods "
            "supported by this revocation endpoint."
        ),
    )

    revocation_endpoint_auth_signing_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the revocation endpoint for the signature on the JWT "
            "used to authenticate the client at the revocation endpoint."
        ),
    )

    introspection_endpoint: str | None = pydantic.Field(
        default=None,
        title="Introspection endpoint",
        description="URL of the authorization server's introspection endpoint.",
    )

    introspection_endpoint_auth_methods_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported authentication methods",
        description=(
            "JSON array containing a list of client authentication methods "
            "supported by this introspection endpoint."
        ),
    )

    introspection_endpoint_auth_signing_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the introspection endpoint for the signature on the JWT "
            "used to authenticate the client at the introspection endpoint."
        ),
    )

    signed_metadata: str | None = pydantic.Field(
        default=None,
        title="Signed metadata",
        description=(
            "Signed JWT containing metadata values about the authorization "
            "server as claims."
        )
    )

    device_authorization_endpoint: str | None = pydantic.Field(
        default=None,
        title="Device authorization endpoint",
        description=(
            "URL of the authorization server's device authorization endpoint."
        ),
    )

    tls_client_certificate_bound_access_tokens: bool | None = pydantic.Field(
        default=True,
        title="Supports mTLS certificate-bound access tokens",
        description=(
            "Indicates authorization server support for mutual-TLS client "
            "certificate-bound access tokens."
        ),
    )

    mtls_endpoint_aliases: dict[str, str] | None = pydantic.Field(
        default={},
        title="Alternative mTLS endpoints",
        description=(
            "JSON object containing alternative authorization server "
            "endpoints, which a client intending to do mutual TLS will "
            "use in preference to the conventional endpoints."
        ),
    )

    nfv_token_signing_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the server for signing the JWT used as NFV Token."
        ),
    )

    nfv_token_encryption_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`alg` values) supported by the server to encode the JWT used as "
            "NFV Token."
        ),
    )

    nfv_token_encryption_enc_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported content encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`enc` values) supported by the server to encode the JWT used as "
            "NFV Token."
        ),
    )

    userinfo_endpoint: str | None = pydantic.Field(
        default=None,
        title="UserInfo endpoint",
        description="URL of the authorization servers' UserInfo Endpoint.",
    )

    acr_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported ACR",
        description=(
            "JSON array containing a list of the Authentication Context Class "
            "References that this authorization server supports."
        ),
    )

    subject_types_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported subject types",
        description=(
            "JSON array containing a list of the Subject Identifier types that "
            "this authorization server supports"
        ),
    )

    id_token_signing_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the server for signing the JWT used as ID Token."
        ),
    )

    id_token_encryption_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`alg` values) supported by the server to encode the JWT used as "
            "ID Token."
        ),
    )

    id_token_encryption_enc_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported content encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`enc` values) supported by the server to encode the JWT used as "
            "ID Token."
        ),
    )

    userinfo_signing_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the server for signing the JWT used as UserInfo Endpoint."
        ),
    )

    userinfo_encryption_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`alg` values) supported by the server to encode the JWT used as "
            "UserInfo Endpoint."
        ),
    )

    userinfo_encryption_enc_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported content encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`enc` values) supported by the server to encode the JWT used as "
            "UserInfo Endpoint."
        ),
    )

    request_object_signing_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the server for signing the JWT used as Request Object."
        ),
    )

    request_object_encryption_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`alg` values) supported by the server to encode the JWT used as "
            "Request Object."
        ),
    )

    request_object_encryption_enc_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported content encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`enc` values) supported by the server to encode the JWT used as "
            "Request Object."
        ),
    )

    display_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported display modes",
        description=(
            "JSON array containing a list of the `display` parameter values "
            "that the OpenID Provider supports."
        ),
    )

    claim_types_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported Claims Types",
        description=(
            "JSON array containing a list of the Claims Types "
            "that the OpenID Provider supports."
        )
    )

    claims_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported Claims Types",
        description=(
            "JSON array containing a list of the Claim Names of the Claims "
            "that the OpenID Provider MAY be able to supply values for."
        )
    )

    claims_locales_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported claim locales",
        description=(
            "Languages and scripts supported for values in Claims being "
            "returned, represented as a JSON array of BCP 47."
        ),
    )

    claims_parameter_supported: bool | None = pydantic.Field(
        default=False,
        title="Supports `claims` parameter?",
        description=(
            "Boolean value specifying whether the OP supports use of the "
            "`claims` parameter."
        ),
    )

    request_parameter_supported: bool | None = pydantic.Field(
        default=False,
        title="Supports `request` parameter?",
        description=(
            "Boolean value specifying whether the OP supports use of the "
            "`request` parameter."
        ),
    )

    request_uri_parameter_supported: bool | None = pydantic.Field(
        default=False,
        title="Supports `request_uri` parameter?",
        description=(
            "Boolean value specifying whether the OP supports use of the "
            "`request_uri` parameter."
        ),
    )

    require_request_uri_registration: bool | None = pydantic.Field(
        default=True,
        title="Requires pre-regiration?",
        description=(
            "Boolean value specifying whether the OP requires any `request_uri` "
            "values used to be pre-registered."
        ),
    )

    require_signed_request_object: bool | None = pydantic.Field(
        default=True,
        title="Requires pre-regiration?",
        description=(
            "Indicates where authorization request needs to be protected as "
            "**Request Object** and provided through either `request` or "
            "`request_uri` parameter."
        ),
    )

    pushed_authorization_request_endpoint: str | None = pydantic.Field(
        default=None,
        title="Pushed Authorization Request (PAR) endpoint",
        description=(
            "URL of the authorization server's pushed authorization request "
            "endpoint."
        ),
    )

    require_pushed_authorization_requests: bool | None = pydantic.Field(
        default=False,
        title="Requires PAR?",
        description=(
            "Indicates whether the authorization server accepts authorization "
            "requests only via PAR."
        ),
    )

    introspection_signing_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of algorithms supported by the "
            "authorization server for introspection response signing."
        ),
    )

    introspection_encryption_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported encryption algorithms",
        description=(
            "JSON array containing a list of algorithms supported by the "
            "authorization server for introspection response content key "
            "encryption (`alg` value)."
        ),
    )

    introspection_encryption_enc_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported content encryption algorithms",
        description=(
            "JSON array containing a list of algorithms supported by the "
            "authorization server for introspection response content "
            "encryption (`enc` value)."
        ),
    )

    authorization_response_iss_parameter_supported: bool | None = pydantic.Field(
        default=False,
        title="Supports `iss` parameter in authorization response?",
        description=(
            "Boolean value indicating whether the authorization server "
            "provides the `iss` parameter in the authorization response."
        ),
    )

    authorization_signing_alg_values_supported: list[str] = pydantic.Field(
        default=[],
        title="JARM signature algorithms",
        description=(
            "A JSON array containing a list of the JWS signing algorithms "
            "(`alg` values) supported by the authorization endpoint to "
            "sign the response."
        )
    )

    authorization_encryption_alg_values_supported: list[str] = pydantic.Field(
        default=[],
        title="JARM key wrapping or key agreement algorithms",
        description=(
            "A JSON array containing a list of the JWS signing algorithms "
            "(`alg` values) supported by the authorization endpoint to "
            "encrypt the response."
        )
    )

    authorization_encryption_enc_values_supported: list[str] = pydantic.Field(
        default=[],
        title="JARM encryption algorithms",
        description=(
            "A JSON array containing a list of the JWE encryption algorithms "
            "(`enc` values) supported by the authorization endpoint to "
            "encrypt the response."
        )
    )

    _discovered: bool = pydantic.PrivateAttr(default=False)
    _jwks: JSONWebKeySet = pydantic.PrivateAttr(default_factory=JSONWebKeySet)

    # TODO: These are non-standard properties and we must submit an RFC
    # to register these at IANA.
    #token_endpoint_auth_encryption_alg_values_supported: list[str] | None = []
    #token_endpoint_auth_encryption_enc_values_supported: list[str] | None = []
    #assertion_signing_alg_values_supported: list[str] | None = []
    #assertion_encryption_alg_values_supported: list[str] | None = []
    #assertion_encryption_enc_values_supported: list[str] | None = []
    #required_encrypted_token_endpoint_auth: bool | None = False
    #require_encrypted_assertion: bool | None = False

    @pydantic.model_validator(mode='before')
    def preprocess(
        cls,
        values: dict[str, Any]
    ) -> dict[str, Any]:
        values.setdefault('token_endpoint_auth_methods_supported', [])
        return values

    @pydantic.field_validator('token_endpoint_auth_methods_supported')
    def preprocess_token_endpoint_auth_methods_supported(
        cls,
        value: list[Any] | None 
    ) -> list[Any]:
        if value is None:
            value = []
        return value

    @property
    def jwks(self) -> JSONWebKeySet:
        return self._jwks

    def supports_response_mode(self, response_mode: Any) -> bool:
        """Return a boolean indicating if the server supports the
        given response mode.
        """
        return bool(self.response_modes_supported)\
            and response_mode in self.response_modes_supported

    @utils.http
    async def discover(self, http: httpx.AsyncClient):
        if self.jwks_uri and not self._discovered:
            response = await http.get(self.jwks_uri, follow_redirects=True)
            self._jwks = JSONWebKeySet.model_validate(response.json())
            self._discovered = True

    @utils.http
    async def userinfo(self, http: httpx.AsyncClient, access_token: str) -> UserInfoResponse:
        if not self.userinfo_endpoint:
            raise NotImplementedError(
                f"Issuer does not implement the UserInfo endpoint."
            )
        response = await http.get(
            url=self.userinfo_endpoint,
            headers={'Authorization': f'Bearer {access_token}'}
        )
        return UserInfoResponse.model_validate_json(response.content)

    async def _discover(self):
        return await self.discover()

    def __await__(self):
        return self._discover().__await__()


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import ipaddress
import urllib.parse
from typing import Any
from typing import Literal
from typing import TypeVar
from typing import Union

from canonical import Hostname
from pydantic_core import CoreSchema
from pydantic_core import core_schema
from pydantic.json_schema import JsonSchemaValue
from pydantic import GetCoreSchemaHandler
from pydantic import GetJsonSchemaHandler


T = TypeVar('T', bound='RedirectURI')


OOB_URLS: set[str] = {
    "urn:ietf:wg:oauth:2.0:oob",
    "urn:ietf:wg:oauth:2.0:oob:auto"
}

IPV6_LOOPBACK_LITERALS: set[str] = {'::1', '0:0:0:0:0:0:0:1'}


class RedirectURI(str):
    __module__: str = 'headless.ext.oidc.types'
    domain: Hostname | None = None
    url: urllib.parse.ParseResult

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        core_schema: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        json_schema = handler(core_schema)
        json_schema = handler.resolve_ref_schema(json_schema)
        return json_schema

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        return core_schema.no_info_after_validator_function(cls, handler(str))

    def __new__(cls, v: str):
        instance = super().__new__(cls, v)
        instance.url = p = urllib.parse.urlparse(v)
        if not v:
            raise ValueError("The redirect_uri parameter can not be empty.")
        if len(v) > 2048:
            raise ValueError('The redirect_uri parameter is too long to be a valid URL.')
        if v in OOB_URLS:
            raise ValueError('OOB is a security risk.')
        if not p.scheme or not p.netloc:
            raise ValueError('The redirect_uri parameter did not specify a valid URL.')
        if p.query:
            raise ValueError('Query parameters must not be used in the redirect URI.')
        if p.fragment:
            raise ValueError('URL fragment must not be used in the redirect URI.')

        # Check if the value is a valid hostname or an IP address.
        if p.hostname is None:
            raise ValueError("Invalid redirect URI.")
        if p.hostname == 'localhost':
            # Clients should use loopback IP literals rather than the
            # string localhost as described in Section 8.4.2.
            # (OAuth 2.1 draft).
            raise ValueError('local redirect URIs must use loopback IP literals.')
        try:
            instance.domain = Hostname.validate(p.hostname) # type: ignore
            if p.scheme != 'https':
                raise ValueError('the https scheme must be used.')
        except ValueError:
            try:
                ip = ipaddress.ip_address(p.hostname)
                p.port
            except ValueError:
                raise ValueError(f'Redirect URI must hold a valid hostname or IP address: {p.hostname}.')
            else:
                if isinstance(ip, ipaddress.IPv6Address)\
                and not str.startswith(p.netloc, '['):
                    raise ValueError("Invalid redirect URI.")
                if not ip.is_loopback and p.scheme != 'https':
                    raise ValueError('the https scheme must be used.')

        return instance
    
    def can_redirect(self, redirect_uri: Union[str, 'RedirectURI']) -> bool:
        """Return a boolean indicating redirection to `redirect_uri`
        is allowed by this instance.
        """
        # Authorization servers MUST require clients to register their
        # complete redirect URI (including the path component).
        # Authorization servers MUST reject authorization requests
        # that specify a redirect URI that doesn't exactly match
        # one that was registered, with an exception for loopback
        # redirects, where an exact match is required except for
        # the port URI component
        if not isinstance(redirect_uri, RedirectURI):
            redirect_uri = RedirectURI(redirect_uri)
        return all([
            self.url.scheme == redirect_uri.url.scheme,
            (self.url.netloc == redirect_uri.url.netloc)\
                or (self.is_loopback() and redirect_uri.is_loopback()),
            self.url.path == redirect_uri.url.path
        ])

    def redirect(
        self,
        allow_params: bool = False,
        mode: Literal['query', 'fragment'] = 'query',
        **params: Any
    ) -> str:
        """Create a redirect URI with the given params."""
        params = {k: v for k, v in params.items() if v is not None}
        p: list[str] = list(urllib.parse.urlparse(self)) # type: ignore
        p[4] = urllib.parse.urlencode(params, quote_via=urllib.parse.quote)
        return urllib.parse.urlunparse(p)

    def is_loopback(self) -> bool:
        """Return a boolean indicating if the :class:`RedirectURI` point
        to a local loopback address.
        """
        assert self.url.hostname is not None
        try:
            ip = ipaddress.ip_address(self.url.hostname)
            return ip.is_loopback
        except ValueError:
            return False

    def __repr__(self) -> str: # pragma: no cover
        return f'RedirectURI({self})'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi
from aiopki.ext import jose

from oauthx.lib.params import Logger
from oauthx.server.request import Request
from ..models import Session
from .sessionsigner import SessionSigner


__all__: list[str] = ['SessionSigner']


SESSION_COOKIE_NAME: str = 'sessionid'


async def get(
    logger: Logger,
    request: Request,
    signer: SessionSigner,
    encoded: str | None = fastapi.Cookie(
        default=None,
        alias=SESSION_COOKIE_NAME
    ),
) -> Session:
    request.session_cookie = SESSION_COOKIE_NAME
    session = Session.new()
    if encoded:
        try:
            jws = jose.parse(encoded)
            if await jws.verify(signer): # type: ignore
                session = jws.payload(Session.model_validate_json)
        except Exception:
            session = Session.new()
            logger.critical(
                "Session signature verification failed, established new (id: %s)",
                session.id
            )
    else:
        logger.debug("Established session (id: %s)", session.id)
    request.set_session(session)
    return session


RequestSession: TypeAlias =  Annotated[Session, fastapi.Depends(get)]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pytest

import oauthx
from .types import AuthorizeFactory


@pytest.fixture(scope='function')
def client_params() -> dict[str, Any]:
    return {
        'sector_identifier_uri': 'https://127.0.0.1'
    }


@pytest.fixture(scope='function')
def grant_types() -> list[str]:
    return []


@pytest.fixture
def redirect_uris() -> list[str]:
    return []


@pytest.mark.skip("Not implemented")
@pytest.mark.asyncio
async def test_client_must_allow_authorization_code_and_not_redirect(
    client: oauthx.Client,
    authorize_request: AuthorizeFactory
):
    _, params, _ = await authorize_request(
        status_code=403,
        expect_error='unauthorized_client',
        response_type='code'
    )
    assert params.is_error()
    assert params.error == 'unauthorized_client'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import urllib.parse
from typing import Any
from typing import TypeVar

from pydantic_core import CoreSchema
from pydantic_core import core_schema
from pydantic.json_schema import JsonSchemaValue
from pydantic import GetCoreSchemaHandler
from pydantic import GetJsonSchemaHandler


T = TypeVar('T', bound='RequestURI')


class RequestURI(str):
    __module__: str = 'headless.ext.oidc.types'
    _parsed: urllib.parse.ParseResult
    _request_id: str | None

    @property
    def id(self) -> str | None:
        return self._request_id

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        core_schema: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        json_schema = handler(core_schema)
        json_schema = handler.resolve_ref_schema(json_schema)
        return json_schema

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _: Any,
        handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        return core_schema.no_info_after_validator_function(cls, handler(str))

    def __new__(cls, v: str):
        if not v:
            raise ValueError('The request_uri parameter can not be empty.')
        if len(v) > 2048:
            raise ValueError('Provided value is too long to be a URI.')
        p = urllib.parse.urlparse(v)
        if p.scheme in {'http', 'https'} and not p.netloc:
            raise ValueError('Provided value is not a valid URI.')
        if not p.scheme:
            raise ValueError(f"The request_uri parameter must be a valid URN.")
        if p.scheme not in {'urn', 'https'}:
            raise ValueError(f'The request_uri parameter uses an unknown scheme.')
        instance = super().__new__(cls, urllib.parse.urlunparse(p))
        instance._parsed = urllib.parse.urlparse(v)
        instance._request_id = None
        if instance._parsed.scheme == 'urn':
            instance._request_id = str.rsplit(instance, ':', 1)[-1]
        return instance
    
    def is_external(self) -> bool:
        return self._parsed.scheme in {'http', 'https'}

    def __repr__(self) -> str: # pragma: no cover
        return f'RequestURI({self})'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
from typing import Any
from typing import Callable
from typing import Coroutine

import fastapi
import fastapi.routing
import starlette.responses
from canonical.exceptions import ProgrammingError

from oauthx.lib.exceptions import InteractionException


class ClientAPIRoute(fastapi.routing.APIRoute):
    logger: logging.Logger = logging.getLogger('uvicorn')

    def get_route_handler(self) -> Callable[[fastapi.Request], Coroutine[Any, Any, starlette.responses.Response]]:
        handler = super().get_route_handler()

        async def f(request: fastapi.Request) -> starlette.responses.Response:
            request = self.setup_request(request)
            try:
                response = await handler(request)
            except InteractionException as e:
                templates = getattr(request, 'templates', None)
                if templates is None:
                    raise ProgrammingError(
                        "Ensure that the ITemplateService dependency is "
                        "injected."
                    )
                response = await e.render_to_response(templates)
            return await self.process_response(response)

        return f

    async def process_response(self, response: fastapi.Response) -> fastapi.Response:
        response.headers.update({
            'Cache-Control': 'no-cache, no-store, must-revalidate',
            'Pragma': 'no-cache',
            'Expires': '0'
        })
        return response

    def setup_request(self, request: Any) -> Any:
        return request


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import cast
from typing import Annotated
from typing import TypeAlias

import fastapi

from oauthx.lib.exceptions import InvalidRequest
from oauthx.lib.params import Logger
from oauthx.lib.types import RequestURI
from oauthx.server import models
from oauthx.server.params import Storage
from oauthx.server.types import AuthorizationRequestKey
from oauthx.server.types import StopSnooping
from .client import Client
from .query import REQUEST
from .query import REQUEST_URI


__all__: list[str] = [
    'AuthorizationRequest'
]


async def get(
    client: Client,
    logger: Logger,
    storage: Storage,
    request: str | None = REQUEST,
    request_uri: RequestURI | None = REQUEST_URI,
) -> models.AuthorizationRequest | None:
    params = None
    if client.must_push() and not request_uri:
        raise InvalidRequest(
            "The client is configured to accept only pushed authorization "
            "requests."
        )
    if request and request_uri:
        raise InvalidRequest(
            "The request and request_uri parameters are mutually"
            "exclusive."
        )
    if request:
        raise NotImplementedError
    if request_uri:
        assert request_uri.id is not None
        params = await storage.get(AuthorizationRequestKey.fromuri(request_uri))
        if params is None:
            raise StopSnooping
        logger.debug(
            "Retrieving existing authorization request (request: %s)",
            params.request_uri
        )

    return cast(models.AuthorizationRequest, params)


AuthorizationRequest: TypeAlias = Annotated[
    models.AuthorizationRequest | None,
    fastapi.Depends(get)
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import fastapi
import fastapi.params

from oauthx.server.config import ProviderConfig
from oauthx.server.params import ClientAuthorizationState
from oauthx.server.params import CurrentSubject
from oauthx.server.params import ObtainedGrant
from oauthx.server.params import OIDCToken
from oauthx.server.params import ReturnURL
from oauthx.server.params import SubjectLogger
from oauthx.server.params import UpstreamProvider
from oauthx.server.request import Request
from oauthx.server.types import AccountIssues
from ..baserequesthandler import BaseRequestHandler
from .params import AuthorizationRequest
from .params import Client
from .params import ClientDependency


class UpstreamCallbackHandler(BaseRequestHandler):
    __module__: str = 'oauthx.server.handlers'
    dependencies: list[fastapi.params.Depends] = [ClientDependency]
    include_in_schema: bool = False
    name: str = 'oauth2.upstream.callback'
    methods: list[str] = ['GET']
    path: str = '/callback'
    provider: ProviderConfig
    response_description: str = "Proceed to the configured return URL"
    return_url: str
    state: ClientAuthorizationState
    summary: str = "Upstream Redirection Endpoint"

    def setup(
        self,
        *,
        client: Client,
        grant: ObtainedGrant,
        id_token: OIDCToken,
        state: ClientAuthorizationState,
        params: AuthorizationRequest,
        provider: UpstreamProvider,
        return_url: ReturnURL,
        subject: CurrentSubject,
        logger: SubjectLogger,
        **_: Any
    ):
        self.client = client
        self.grant = grant
        self.id_token = id_token
        self.params = params
        self.provider = provider
        self.publisher = logger
        self.return_url = return_url
        self.subject = subject
        self.state = state

    async def handle(self, request: Request) -> fastapi.Response:
        """Retrieve the Open ID Connect ID Token from the identity
        provider and authenticate the user. On success, redirect
        the user-agent to the ``return-url`` that was configured
        when the authorization request was created.
        """
        if self.id_token.email is None:
            raise NotImplementedError
        if self.subject is not None:
            raise NotImplementedError
        try:
            subject, created = await self.server.onboard_oidc(
                token=self.id_token,
                use=self.provider.audience,
                request_id=self.params.id,
                client_id=self.client.client_id,
            )
        except self.server.OrphanedPrincipal:
            raise AccountIssues
        await self.storage.delete(self.state)
        if created:
            self.publisher.onboarded(
                sub=int(subject.get_primary_key()), # type: ignore
                registrar='self',
                authorization_id=self.params.id,
                client_id=self.client.client_id
            )
        request.authenticate(
            sub=str(subject.get_primary_key()),
            email=self.id_token.email
        )
        return self.redirect(self.return_url)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import contextlib
import functools
import inspect
from typing import Any
from typing import AsyncIterable
from typing import Iterable
from typing import TypeVar

import pydantic
from canonical import ResourceIdentifier

from oauthx.lib import utils
from oauthx.lib.models import ClientAuthorizationStateKey
from oauthx.lib.models import ClientAuthorizationState
from oauthx.lib.protocols import IStorage
from oauthx.lib.types import MaskedPrincipal
from oauthx.server import types
from oauthx.server.models import Authorization
from oauthx.server.models import AuthorizationRequest
from oauthx.server.models import Client
from oauthx.server.models import Principal
from oauthx.server.models import Receipt
from oauthx.server.models import ResourceOwner
from oauthx.server.models import Subject
from oauthx.server.models import SubjectKey
from oauthx.server.types import AuthorizationKey
from oauthx.server.types import AuthorizationRequestKey
from oauthx.server.protocols import IResourceOwner
from oauthx.server.protocols import ISubject
from .config import Config
from .localclientstorage import LocalClientStorage
from .params import CurrentConfig


T = TypeVar('T')


class BaseStorage(IStorage):
    __module__: str = 'oauthx.server.ref'
    _client: Client | None
    clients: LocalClientStorage
    config: Config

    @utils.class_property
    def __signature__(cls) -> inspect.Signature:
        return utils.merge_signatures([
            inspect.signature(cls.__init__),
            inspect.signature(cls.setup)
        ])

    def __init__(self, config: Config = CurrentConfig, **kwargs: Any) -> None:
        self._client = config.client
        self.config = config
        self.clients = LocalClientStorage()
        self.setup(**kwargs)

    def setup(self, **kwargs: Any) -> None:
        pass

    def find(
        self,
        kind: type[pydantic.BaseModel],
        filters: Iterable[tuple[str, str, str]],
        sort: Iterable[str] | None = None
    ) -> AsyncIterable[pydantic.BaseModel]:
        raise NotImplementedError

    @contextlib.asynccontextmanager
    async def atomic(self, key: Any):
        obj = await self.get(key)
        try:
            yield obj
            await self.persist(obj)
        except Exception:
            raise

    async def allocate_identifier(self, cls: str | type) -> Any:
        raise NotImplementedError

    async def delete(self, obj: Any) -> None:
        raise NotImplementedError

    @functools.singledispatchmethod
    async def _get(self, key: Any) -> Any:
        if key is None:
            return None
        raise NotImplementedError(type(key).__name__)

    @_get.register
    async def _(self, key: AuthorizationKey) -> Authorization | None:
        return await self.get_authorization(key)

    @_get.register
    async def _(self, key: AuthorizationRequestKey) -> types.IAuthorizationRequest | None:
        return await self.get_authorization_request(key)

    @_get.register
    async def _(self, key: Client.Key) -> Client | None:
        if self._client and self._client.id == key:
            return self._client
        return await self.clients.get(str(key)) or await self.get_client(key)

    @_get.register
    async def _(self, key: ClientAuthorizationStateKey) -> ClientAuthorizationState | None:
        return await self.get_authorization_state(key)

    @_get.register
    async def _(self, masked: MaskedPrincipal) -> Principal | None:
        return await self.get_principal(masked)

    @_get.register
    async def _(self, key: ResourceOwner.Key) -> IResourceOwner | None:
        return await self.get_resource_owner(key)

    @_get.register
    async def _(self, pk: SubjectKey) -> ISubject | None:
        return await self.get_subject_by_pk(pk)

    async def get(self, key: ResourceIdentifier[Any, T]) -> T | None:
        try:
            return await self._get(key)
        except LookupError:
            return None

    async def get_authorization(self, key: AuthorizationKey) -> Authorization | None:
        raise NotImplementedError

    async def get_authorization_request(self, key: AuthorizationRequestKey) -> types.IAuthorizationRequest | None:
        raise NotImplementedError

    async def get_authorization_state(self, key: ClientAuthorizationStateKey) -> Any:
        raise NotImplementedError

    async def get_client(self, key: Client.Key) -> Client | None:
        raise NotImplementedError

    async def get_principal(self, masked: MaskedPrincipal) -> Principal | None:
        raise NotImplementedError

    async def get_resource_owner(self, key: ResourceOwner.Key) -> IResourceOwner | None:
        raise NotImplementedError

    async def get_subject_by_pk(self, pk: SubjectKey) -> ISubject | None:
        raise NotImplementedError

    async def get_subject_by_principal(self, masked: MaskedPrincipal) -> ISubject | None:
        raise NotImplementedError

    async def mask(self, obj: Any) -> Any:
        raise NotImplementedError

    async def persist(self, object: Any) -> None:
        return await self._persist(object)

    async def persist_authorization(self, authorization: Authorization) -> None:
        raise NotImplementedError

    async def persist_authorization_request(self, request: AuthorizationRequest) -> None:
        raise NotImplementedError

    async def persist_authorization_state(
            self,
            state: ClientAuthorizationState
    ) -> None:
        raise NotImplementedError

    async def persist_client(self, client: Client) -> None:
        raise NotImplementedError

    async def persist_receipt(self, receipt: Receipt) -> None:
        raise NotImplementedError

    async def persist_resource_owner(self, owner: ResourceOwner) -> None:
        raise NotImplementedError

    async def persist_principal(self, principal: Principal) -> None:
        raise NotImplementedError

    async def persist_subject(self, subject: Subject) -> None:
        raise NotImplementedError

    @functools.singledispatchmethod
    async def _persist(self, object: Any) -> None:
        raise NotImplementedError

    @_persist.register
    async def _(self, authorization: Authorization) -> None:
        return await self.persist_authorization(authorization)

    @_persist.register
    async def _(self, request: AuthorizationRequest) -> None:
        return await self.persist_authorization_request(request)

    @_persist.register
    async def _(self, client: Client) -> None:
        return await self.persist_client(client)

    @_persist.register
    async def _(self, state: ClientAuthorizationState) -> None:
        return await self.persist_authorization_state(state)

    @_persist.register
    async def _(self, principal: Principal) -> None:
        return await self.persist_principal(principal)

    @_persist.register
    async def _(self, receipt: Receipt) -> None:
        return await self.persist_receipt(receipt)

    @_persist.register
    async def _(self, owner: ResourceOwner) -> None:
        return await self.persist_resource_owner(owner)

    @_persist.register
    async def _(self, subject: Subject) -> None:
        return await self.persist_subject(subject)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
from typing import Any
from typing import Iterable
from typing import Literal
from typing import MutableMapping
from typing import TypeAlias

import pydantic

from oauthx.lib.protocols import IStorage
from oauthx.server.protocols import IUserInfoContributor
from oauthx.server.types import AuthorizationKey
from .subjectkey import SubjectKey


AuthorizedTokenType: TypeAlias = Literal[
    'urn:ietf:params:oauth:token-type:access_token',
    'urn:ietf:params:oauth:token-type:refresh_token',
    'urn:ietf:params:oauth:token-type:id_token',
]


class Authorization(pydantic.BaseModel):
    """A :class:`Authorization` object represents the authorization granted
    by the resource owner through the authorization endpoint.
    """
    authorized: datetime.datetime = pydantic.Field(
        default_factory=lambda: datetime.datetime.now(datetime.timezone.utc)
    )

    client_id: str = pydantic.Field(
        default=...
    )

    id: int = pydantic.Field(
        default=...
    )
    
    issuer: str = pydantic.Field(
        default=...
    )

    lifecycle: Literal['GRANTED', 'ISSUED'] = pydantic.Field(
        default='ISSUED'
    )

    prompted: bool = pydantic.Field(
        default=False
    )

    scope: set[str] = pydantic.Field(
        default_factory=set
    )

    sub: SubjectKey = pydantic.Field(
        default=...
    )
    
    token_types: list[AuthorizedTokenType] = pydantic.Field(
        default=...
    )
    
    userinfo: dict[str, Any] = pydantic.Field(
        default={}
    )

    @property
    def pk(self) -> AuthorizationKey:
        return AuthorizationKey(self.id)

    def allows_scope(self, scope: Iterable[str]) -> bool:
        return set(scope) <= self.scope

    def is_consumed(self) -> bool:
        return self.lifecycle == 'ISSUED'

    def contribute(self, contributor: IUserInfoContributor) -> None:
        contributor.contribute_to_userinfo(self.userinfo)

    def contribute_to_userinfo(self, userinfo: MutableMapping[str, Any]) -> None:
        userinfo.update(self.userinfo)

    def grants_id_token(self) -> bool:
        return 'urn:ietf:params:oauth:token-type:id_token' in self.token_types

    async def consume(self, storage: IStorage) -> None:
        self.lifecycle = 'ISSUED'
        await storage.persist(self)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import contextlib
from typing import Any
from typing import AsyncGenerator
from typing import AsyncIterable
from typing import Iterable
from typing import Protocol
from typing import TypeAlias
from typing import TypeVar

from canonical import ResourceIdentifier
import pydantic

from oauthx.lib.types import Masked


P = TypeVar('P', bound=pydantic.BaseModel)
T = TypeVar('T')


class IStorage(Protocol):
    __module__: str = 'oauthx.lib.protocols'
    Findable: TypeAlias = type[P]

    def find(
        self,
        kind: type[P],
        filters: Iterable[tuple[str, str, Any]],
        sort: Iterable[str] | None = None
    ) -> AsyncIterable[P]:
        """Find objects by the given filters."""
        ...

    async def allocate_identifier(self, cls: str | type) -> Any: ...
    async def delete(self, obj: Any) -> None: ...
    async def get(self, key: ResourceIdentifier[Any, T]) -> T | None: ...
    async def mask(self, obj: Any) -> Masked: ...
    async def persist(self, object: Any) -> None: ...

    @contextlib.asynccontextmanager
    async def atomic(self, key: ResourceIdentifier[Any, T]) -> AsyncGenerator[T | None, None]:
        obj = await self.get(key)
        try:
            yield obj
            await self.persist(obj)
        except Exception:
            pass


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import uuid
from typing import Any

import fastapi
from canonical.protocols import ITemplateService


class InteractionException(Exception):
    """Base class for all fatal and non-redirectable exceptions
    raised during an interaction with a :term:`Resource Owner`.
    """
    __module__: str = 'oauthx.server.types'
    allow_redirect: bool = True
    error: str = 'illegal_request'
    error_description: str = 'Your request is blocked.'
    media_type: str = 'text/html'
    status_code: int = 400
    template_name: str = NotImplemented

    def __init__(self, context: dict[str, Any] | None = None) -> None:
        self.context = context or {}
        self.id = str(uuid.uuid4())

    def get_context(self) -> dict[str, Any]:
        return {**self.context, 'incident_id': self.id}

    def get_template_names(self) -> list[str]:
        """Return the list of template names to render the HTML response
        for this exception.
        """
        if self.template_name == NotImplemented:
            raise NotImplementedError
        return [self.template_name]

    async def render_to_response(self, templates: ITemplateService) -> fastapi.Response:
        return fastapi.responses.HTMLResponse(
            status_code=self.status_code,
            headers={
                'X-Error': self.error,
                'X-Error-Description': self.error_description
            },
            content=await templates.render_template(
                templates=self.get_template_names(),
                context=self.get_context()
            )
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi
import pydantic

from oauthx.lib.exceptions import InvalidAuthorizationResponse
from oauthx.lib.models import AuthorizationResponse as AuthorizationResponseModel


async def get(
    error: str | None = fastapi.Query(
        default=None,
        description=(
            "Error code returned by the authorization "
            "server."
        )
    ),
    error_description: str | None = fastapi.Query(
        default=None,
        title="Error description"
    ),
    code: str | None = fastapi.Query(
        default=None,
        title="Authorization code",
        description=(
            "The authorization code supplied by the server."
        )
    ),
    state: str | None = fastapi.Query(
        default=None,
        title="State",
        description=(
            "The `state` parameter that was included with the authorization "
            "request."
        )
    ),
    iss: str | None = fastapi.Query(
        default=None,
        title="Issuer identifier",
        description=(
            "The issuer identifier that clients must use to verify the "
            "response of the authorization endpoint."
        )
    )
) -> AuthorizationResponseModel:
    try:
        return AuthorizationResponseModel.model_validate({
            'code': code,
            'error': error,
            'error_description': error_description,
            'state': state,
            'iss': iss
        })
    except pydantic.ValidationError:
        raise InvalidAuthorizationResponse(error='invalid_response')


AuthorizationResponse: TypeAlias = Annotated[
    AuthorizationResponseModel,
    fastapi.Depends(get)
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import functools
from typing import Any
from typing import AsyncIterable
from typing import Callable
from typing import Iterable
from typing import TypeVar

import pydantic
from canonical import ResourceIdentifier

from oauthx.lib.models import ClientAuthorizationState
from oauthx.lib.models import ClientAuthorizationStateKey
from oauthx.lib.types import MaskedPrincipal
from oauthx.server import types
from oauthx.server import BaseStorage
from oauthx.server.types import AuthorizationKey
from oauthx.server.models import Authorization
from oauthx.server.models import AuthorizationRequest
from oauthx.server.models import Claim
from oauthx.server.models import ClaimSet
from oauthx.server.models import Principal
from oauthx.server.models import Receipt
from oauthx.server.models import ResourceOwner
from oauthx.server.models import ResourceOwnerKey
from oauthx.server.models import Subject
from oauthx.server.models import SubjectKey
from oauthx.server.params import ContentEncryptionKey
from .params import DatastoreClient
from .types import IDatastoreKey
from .types import IDatastoreEntity


T = TypeVar('T', bound=pydantic.BaseModel)


class DatastoreStorage(BaseStorage):

    @staticmethod
    async def run_in_executor(
        func: Callable[..., Any],
        *args: Any,
        **kwargs: Any
    ) -> Any:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, func, *args, **kwargs)

    def setup( # type: ignore
        self,
        *,
        client: DatastoreClient, # type: ignore
        key: ContentEncryptionKey,
        **kwargs: Any
    ) -> None:
        self.client = client
        self.key = key

    async def allocate_identifier(self, cls: type | str) -> int:
        if not isinstance(cls, str):
            cls = cls.__name__
        base = self.entity_key(kind=cls)
        result = await self.run_in_executor(
            functools.partial( # type: ignore
                self.client.allocate_ids, # type: ignore
                incomplete_key=base,
                num_ids=1
            )
        )
        return [x for x in result][0].id

    async def delete(self, obj: Any) -> None:
        key = self.entity_key(type(obj), obj.pk)
        await self.run_in_executor(
            functools.partial(
                self.client.delete, # type: ignore
                key=key
            )
        )

    def entity_factory(self, key: IDatastoreKey, obj: pydantic.BaseModel) -> IDatastoreEntity:
        entity = self.client.entity(key) # type: ignore
        attrs = obj.model_dump(mode='json')
        if attrs:
            entity.update(attrs) # type: ignore
        return entity # type: ignore

    def entity_key(
        self,
        kind: str | type,
        identifier: int | str | ResourceIdentifier[int | str, Any] | None = None,
        parent: IDatastoreKey | None = None
    ) -> IDatastoreKey:
        if not isinstance(kind, str):
            kind = kind.__name__
        if isinstance(identifier, ResourceIdentifier):
            identifier = identifier.cast() # type: ignore
        args: list[Any] = [kind]
        if identifier is not None:
            args.append(identifier)
        return self.client.key(*args, parent=parent) # type: ignore

    async def find(
        self,
        kind: type[T],
        filters: Iterable[tuple[str, str, int | str]],
        sort: Iterable[str] | None = None
    ) -> AsyncIterable[T]:
        q = self.client.query(kind=kind.__name__) # type: ignore
        if sort is not None:
            q.order = sort
        for filter in filters:
            q.add_filter(*filter) # type: ignore
        results = await self.run_in_executor(
            functools.partial(q.fetch) # type: ignore
        )
        for entity in results:
            yield kind.model_validate(dict(entity))

    async def get_entity_by_key(self, key: IDatastoreKey) -> IDatastoreEntity | None:
        return await self.run_in_executor(
            functools.partial(
                self.client.get, # type: ignore
                key=key
            )
        )

    async def get_model_by_key(self, cls: type[T], pk: int | str | ResourceIdentifier[Any, Any]) -> T | None:
        entity = await self.get_entity_by_key(
            key=self.entity_key(cls, pk)
        )
        if entity is None:
            return None
        return cls.model_validate(dict(entity))

    async def put(self, obj: pydantic.BaseModel, pk: int | str | ResourceIdentifier[Any, Any]) -> IDatastoreKey:
        key = self.entity_key(type(obj), pk)
        entity = self.entity_factory(key, obj)
        await self.run_in_executor(self.client.put, entity) # type: ignore
        return entity.key # type: ignore

    async def get_authorization(self, key: AuthorizationKey) -> Authorization | None:
        return await self.get_model_by_key(Authorization, int(key))

    async def get_authorization_request(self, key: types.AuthorizationRequestKey) -> types.IAuthorizationRequest | None:
        return await self.get_model_by_key(AuthorizationRequest, int(key)) # type: ignore

    async def get_authorization_state(self, key: ClientAuthorizationStateKey) -> ClientAuthorizationState | None:
        return await self.get_model_by_key(ClientAuthorizationState, key)

    async def get_principal(self, masked: MaskedPrincipal) -> Principal | None:
        return await self.get_model_by_key(Principal, str(masked))

    async def get_resource_owner(self, key: ResourceOwnerKey) -> ResourceOwner | None:
        return await self.get_model_by_key(ResourceOwner, str(key))

    async def get_subject_by_pk(self, pk: SubjectKey) -> Subject | None:
        return await self.get_model_by_key(Subject, int(pk))

    async def persist_authorization(self, authorization: Authorization) -> None:
        await self.put(authorization, int(authorization.pk))

    async def persist_authorization_request(self, request: AuthorizationRequest) -> None:
        await self.put(request, request.id)

    async def persist_authorization_state(
            self,
            state: ClientAuthorizationState
    ) -> None:
        await self.put(state, state.pk)

    async def persist_principal(self, principal: Principal) -> None:
        if not principal.is_encrypted():
            raise ValueError("Call Principal.encrypt(subject) prior to persisting.")
        await self.put(principal, str(principal.masked))

    async def persist_receipt(self, receipt: Receipt) -> None:
        claims = await ClaimSet.fromquery(
            provider=receipt.provider,
            query=self.find(
                kind=Claim,
                filters=[
                    ('provider', '=', str(receipt.provider)),
                    ('sub', '=', receipt.sub)
                ]
            )
        )
        for claim in claims.diff(receipt.claims):
            await self.put(claim, claim.pk)

    async def persist_resource_owner(self, owner: ResourceOwner) -> None:
        await self.put(owner, str(owner.pk))

    async def persist_subject(self, subject: Subject) -> None:
        if not subject.is_encrypted():
            raise ValueError("Call Subject.encrypt_keys() prior to persisting.")
        await self.put(subject, int(subject.get_primary_key()))


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias
import fastapi

from oauthx.lib.params import Logger
from oauthx.server.protocols import ISubject
from oauthx.server.request import Request
from .storage import Storage


__all__: list[str] = ['CurrentSubject']


async def get(
    logger: Logger,
    request: Request,
    storage: Storage
) -> ISubject | None:
    if request.subject:
        logger.debug(
            "Retrieving Subject from storage (sub: %s)",
            str(request.subject.pk)
        )
        return await storage.get(request.subject.pk)


CurrentSubject: TypeAlias =  Annotated[ISubject | None, fastapi.Depends(get, use_cache=True)]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import ssl
from typing import cast
from typing import Annotated
from typing import Any
from typing import AsyncGenerator
from typing import Callable
from typing import Literal
from typing import TypeVar

import httpx
import fastapi
import fastapi.params
from fastapi.security import HTTPBearer
from fastapi.security import HTTPAuthorizationCredentials
from aiopki.ext import jose

from oauthx.lib import LazyServerMetadata
from oauthx.lib import RFC9068AccessToken
from oauthx.lib import ServerMetadata
from oauthx.lib.exceptions import InvalidRequest
from oauthx.lib.exceptions import InvalidToken
from oauthx.lib.types import AccessTokenHash


__all__: list[str] = ['RequestAccessToken']

T = TypeVar('T', bound=RFC9068AccessToken)

ISSUER_CACHE: dict[str, ServerMetadata | LazyServerMetadata] = {}


async def get_http_client() -> AsyncGenerator[httpx.AsyncClient, None]:
    async with httpx.AsyncClient() as client:
        yield client


async def get_issuer_metadata(
    http: httpx.AsyncClient,
    issuer: str,
    timeout: int = 5
) -> ServerMetadata:
    ISSUER_CACHE.setdefault(issuer, LazyServerMetadata(iss=issuer))
    metadata: ServerMetadata | LazyServerMetadata = ISSUER_CACHE[issuer]
    if isinstance(metadata, LazyServerMetadata):
        try:
            metadata = await metadata.discover(http=http)
        except (asyncio.TimeoutError, asyncio.CancelledError, httpx.ReadTimeout):
            raise InvalidToken("The issuer did not respond in time.")
        except ssl.SSLCertVerificationError:
            raise InvalidToken(
                "Unable to establish a secure connection with the issuer "
                "of the access token."
            )
        except Exception:
            raise InvalidToken(
                "The issuer of the access token is not a "
                "discoverable authorization server."
            )
        ISSUER_CACHE[issuer] = metadata
    assert isinstance(metadata, ServerMetadata)
    return metadata


def RequestAccessToken(
    issuers: str | set[str] | None,
    factory:  Callable[[bytes], T] = RFC9068AccessToken.model_validate,
    max_age: int | None = None,
    issuer_timeout: int = 5,
    required: bool = True,
    audience: Literal['server', 'endpoint', 'any'] = 'server',
    types: set[str | None] = {'at+jwt', 'application/at+jwt'},
    http_factory: Any = get_http_client
) -> fastapi.params.Depends:
    issuers = issuers or set()
    if isinstance(issuers, str):
        issuers = {issuers}
    if not isinstance(http_factory, fastapi.params.Depends):
        http_factory = fastapi.Depends(http_factory)


    async def get(
        request: fastapi.Request,
        authorization: Annotated[
            HTTPAuthorizationCredentials | None,
            fastapi.Depends(HTTPBearer(auto_error=False))
        ],
        http: httpx.AsyncClient = cast(Any, http_factory)
    ) -> RFC9068AccessToken | None:
        trust = set(issuers)
        if 'self' in trust:
            trust.remove('self')
            trust.add(f'{request.url.scheme}://{request.url.netloc}')
        allow: set[str] = {
            f'{request.url.scheme}://{request.url.netloc}{request.url.path}',
        }
        if audience == 'server':
            allow = {f'{request.url.scheme}://{request.url.netloc}'}
        if authorization is None:
            if required:
                raise InvalidRequest("The Authorization header is required.")
            return None
        if str.lower(authorization.scheme) != 'bearer':
            raise InvalidRequest("The 'Bearer' scheme is required.")
        try:
            jws = jose.parse(authorization.credentials)
            at = jws.payload(factory=factory)
        except Exception:
            raise InvalidRequest("Malformed access token.")
        if jws.is_encrypted():
            raise InvalidRequest("Encrypted access tokens are not supported.")
        if jws.typ not in types:
            raise InvalidToken("Token 'typ' claim must be at+jwt or application/at+jwt.")
        if audience != 'any' and not at.validate_aud(allow):
            raise InvalidToken("This resource server is not the audience of the access token.")
        if not at.validate_exp():
            raise InvalidToken("The access token is expired.")
        if not any(map(at.validate_iss, trust)):
            raise InvalidToken(
                "The access token is issued by an untrusted authorization server. "
                f"Trusted authorization servers are: {', '.join(sorted(trust))}"
            )
        if not at.validate_nbf(required=False):
            raise InvalidToken("The access token is inactive.")
        if max_age is not None and not at.validate_iat(max_age=max_age, required=True):
            raise InvalidToken("The access token is too old.")
        assert at.iss is not None
        metadata = await get_issuer_metadata(
            http=http,
            issuer=at.iss,
            timeout=issuer_timeout
        )
        if not await jws.verify(metadata.jwks):
            raise InvalidToken("The signature of the access token did not validate.")
        setattr(request, 'access_token', at)
        setattr(request, 'credentials', authorization.credentials)
        setattr(request, 'at_hash', AccessTokenHash.parse_at(authorization.credentials))
        setattr(request, 'issuer', metadata)

    return fastapi.Depends(get)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
import uuid
from typing import Any
from typing import Callable
from typing import Coroutine

import fastapi
import fastapi.routing
import starlette.responses

from oauthx.lib.exceptions import Error
from oauthx.lib.exceptions import UserAgentException
from .config import Config
from .request import Request
from .types import InteractionException
from .types import InvalidAuthorizationRequest
from .types import LoginRequired


class OIDCRouteHandler(fastapi.routing.APIRoute):
    config: Config
    logger: logging.Logger = logging.getLogger('uvicorn')
    login_endpoint: str

    def get_route_handler(self) -> Callable[[fastapi.Request], Coroutine[Any, Any, starlette.responses.Response]]:
        handler = super().get_route_handler()

        async def f(request: fastapi.Request) -> starlette.responses.Response:
            request = Request(request.scope, request.receive)
            request.config = self.config # type: ignore
            try:
                response = await handler(request)
            except fastapi.exceptions.RequestValidationError as e:
                # If any of the error is in the path, then the response is
                # 404, else 400.
                raise
                raise NotImplementedError
            except UserAgentException as e:
                response = await e.render_to_response(
                    request=request,
                    templates=request.templates
                )
            except LoginRequired as e:
                response = e.redirect(request, self.login_endpoint, str(request.url))
            except (Error, InteractionException, InvalidAuthorizationRequest) as e:
                self.logger.error(
                    "Caught fatal %s (error: %s, error_description: %s)",
                    type(e).__name__, e.error, e.error_description)
                response = await request.error(
                    e.get_template_names(),
                    error=e.error,
                    error_description=e.error_description,
                    status_code=e.status_code,
                    media_type=e.media_type,
                    context={**e.get_context(), 'exception': e},
                    allow_redirect=e.allow_redirect
                )
            except Exception as e:
                self.logger.exception("Caught fatal %s", type(e).__name__)
                response = await request.error(
                    'oauthx/errors/fatal.html.j2',
                    error='server_error',
                    error_description=(
                        "The authorization server encountered an unexpected "
                        "condition that prevented it from fulfilling the "
                        "request."
                    ),
                    status_code=500,
                    context={'incident_id': str(uuid.uuid4())}
                )
            response = await request.process_response(response)
            response.headers.update({
                'permissions-policy': str.join(',', [
                    'accelerometer=()',
                    'ambient-light-sensor=()',
                    'autoplay=()',
                    'battery=()',
                    'camera=()',
                    'display-capture=()',
                    'document-domain=()',
                    'encrypted-media=()',
                    'fullscreen=()',
                    'gamepad=()',
                    'geolocation=()',
                    'gyroscope=()',
                    'layout-animations=(self)',
                    'legacy-image-formats=(self)',
                    'magnetometer=()',
                    'microphone=()',
                    'midi=()',
                    'oversized-images=(self)',
                    'payment=()',
                    'picture-in-picture=()',
                    'publickey-credentials-get=()',
                    'speaker-selection=()',
                    'sync-xhr=(self)',
                    'unoptimized-images=(self)',
                    'unsized-media=(self)',
                    'usb=()',
                    'screen-wake-lock=()',
                    'web-share=()',
                    'xr-spatial-tracking=()'
                ]),
                'referrer-policy': 'no-referrer',
                'x-content-type-options': 'nosniff',
                'x-frame-options': 'deny',
                'x-permitted-cross-domain-policies': 'none',
            })
            if request.url.scheme == 'https':
                response.headers['strict-transport-security'] = 'max-age=31536000; includeSubDomains; preload'
            return response

        return f


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import json
from typing import Any
from typing import Callable

import fastapi
import fastapi.params
import httpx
import yaml
from aiopki.ext.jose import JWA
from aiopki.ext.jose import JWKS
from aiopki import CryptoKeyType
from aiopki.types import ISigner
from jinja2 import Environment
from jinja2 import BaseLoader
from jinja2 import PackageLoader
from jinja2 import ChoiceLoader
from jinja2 import FileSystemLoader

from oauthx.lib.models import ServerMetadata
from oauthx.lib.params import ApplicationStorage
from oauthx.lib.templates import TemplateService
from oauthx.lib.utils import set_signature_defaults
from .authorizationserver import AuthorizationServer
from .config import Config
from .handlers import AuthorizationRequestHandler
from .handlers import TokenEndpointHandler
from .handlers import UpstreamBeginHandler
from .handlers import UpstreamCallbackHandler
from .handlers import UserInfoEndpointHandler
from .clientroutehandler import ClientRouteHandler
from .oidcroutehandler import OIDCRouteHandler
from .params import PendingAuthorizationClient
from .params import PendingAuthorizationRequest
from .params import Storage
from .pluginendpoint import PluginEndpoint
from .request import Request
from .subjectlogger import SubjectLogger


class OIDCRouter(fastapi.APIRouter):
    __module__: str = 'oauthx.server'
    config: Config
    http: httpx.AsyncClient | None = None
    jwks: JWKS
    signer: ISigner | None = None

    @classmethod
    def create_server(
        cls,
        config: str | Config,
        dependencies: list[Any] = [],
        asgi_class: type[fastapi.FastAPI] = fastapi.FastAPI,
        http: httpx.AsyncClient | None = None,
        **kwargs: Any
    ):
        if not isinstance(config, Config):
            config = Config.load(config)
        app = asgi_class(dependencies=dependencies, **kwargs)
        router = cls(
            app=app,
            config=config,
            http=http,
        )
        app.include_router(
            router=router
        )
        return app

    @classmethod
    def from_config(cls, app: fastapi.FastAPI, config: str | Config, **kwargs: Any):
        if not isinstance(config, Config):
            config = Config.model_validate({
                **yaml.safe_load(open(config).read())
            })
        return cls(
            app=app,
            config=config,
            **kwargs
        )

    def __init__(
        self,
        *,
        app: fastapi.FastAPI,
        config: Config,
        template_service: type[TemplateService] = TemplateService,
        authorize_handler: type[AuthorizationRequestHandler] = AuthorizationRequestHandler,
        token_endpoint_handler: type[TokenEndpointHandler] = TokenEndpointHandler,
        login_endpoint: str = 'oauth2.login',
        login_path: str = '/login',
        prefix: str = '/oauth/v2',
        http: httpx.AsyncClient | None = None
    ):
        super().__init__(
            dependencies=[
                fastapi.Depends(SubjectLogger),
                self._wrap_depends('config', self.get_config),
                ApplicationStorage(config.storage.impl.value),
                self._wrap_depends('http', lambda: self.http),
                self._wrap_depends('keychain', self.get_keychain),
                self._wrap_depends('plugins', lambda: self.plugins),
                self._wrap_depends('signer', self.get_signer),
                self._wrap_depends('token_signer', lambda: self.signing_key),
                self._wrap_depends('templates', lambda: template_service(self.templates)),
                self._wrap_depends('oauth', AuthorizationServer),
            ],
            route_class=type('OIDCRouteHandler', (OIDCRouteHandler,), {
                'config': config,
                'login_endpoint': login_endpoint
            }),
        )
        self.app = app
        self.config = config
        self.http = http
        self.plugins = [plugin.value for plugin in config.plugins]
        self.signing_key = config.issuer.signing_key
        self.subject_class = config.authentication.impl.value
        self.template_loaders: list[BaseLoader] = [
            FileSystemLoader('templates/oauthx'),
            PackageLoader('oauthx.lib')
        ]

        # RFC 8414 OAuth 2.0 Authorization Server Metadata
        @app.get('/.well-known/openid-configuration')
        def _(request: Request) -> fastapi.Response:
            return fastapi.responses.RedirectResponse(
                status_code=301,
                url=request.url_for('oauth2.metadata')
            )

        app.add_api_route(
            methods=['GET'],
            path=f'/.well-known/oauth-authorization-server',
            endpoint=self.metadata,
            name='oauth2.metadata',
            summary='Metadata Endpoint',
            response_model=ServerMetadata,
            tags=['OAuth 2.x/OpenID Connect']
        )

        app.add_api_route(
            methods=['GET'],
            path='/.well-known/jwks.json',
            endpoint=self.keys,
            name='oauth2.jwks',
            summary='JSON Web Key Set (JWKS)',
            response_model=JWKS,
            tags=['OAuth 2.x/OpenID Connect']
        )

        # Add the login handler to the fastapi.FastAPI instance instead
        # of the router, as the implementer possibly does not want to
        # expose the login endpoint under /oauth/v2.
        for handler in config.authentication.handlers:
            handler.impl.value.add_to_router(
                router=self,
                name=handler.name,
                path=handler.path,
                dependencies=[
                    self._wrap_depends('subject', self.subject_class.resolve)
                ],
            )

        # If there are upstream identity providers trusted by the
        # authorization server, then add UpstreamLoginRequestHandler
        # so that frontend clients can initate the login flow. This
        # endpoint is added to the router as it is an OAuth-specific
        # concern.

        self.add_api_route(
            methods=['POST'],
            path=f'{prefix}/token',
            endpoint=set_signature_defaults(self.token, {
                'handler': fastapi.Depends(token_endpoint_handler)
            }),
            name='oauth2.token',
            summary="Token Endpoint",
            tags=['OAuth 2.x/OpenID Connect'],
            route_class_override=type('ClientRouteHandler', (ClientRouteHandler,), {
                'config': config,
                'login_endpoint': login_endpoint
            }),
        )

        if config.has_authorize_endpoint():
            authorize_handler.add_to_router(
                router=self,
                dependencies=[
                    self._wrap_depends('subject', self.subject_class.resolve)
                ],
                prefix=prefix
            )

        if self.config.providers:
            UpstreamBeginHandler.add_to_router(
                router=self,
                dependencies=[
                    self._wrap_depends('subject', self.subject_class.resolve)
                ],
                prefix=prefix
            )
            UpstreamCallbackHandler.add_to_router(
                router=self,
                dependencies=[
                    self._wrap_depends('subject', self.subject_class.resolve)
                ],
                prefix=prefix
            )

        UserInfoEndpointHandler.add_to_router(self, prefix=prefix)

        # User interface endpoints
        self.add_api_route(
            methods=['GET'],
            path='/a/{pk}/deny',
            endpoint=self.deny,
            status_code=302,
            name='oauth2.deny',
            summary="Deny authorization request",
            include_in_schema=False
        )

        # Load plugins. This should run before initializing the template
        # loaders because plugins are allowed to add them using the
        # OIDCRouter.register_template_module() method.
        for plugin in self.plugins:
            plugin.__register__(self)

        # Initialize templates
        self.templates = Environment(
            extensions=[
                'jinja2.ext.i18n',
                'jinja_markdown.MarkdownExtension'
            ],
            loader=ChoiceLoader(self.template_loaders)
        )
        self.templates.install_null_translations() # type: ignore

    def add_plugin(
        self,
        *,
        methods: list[str],
        path: str,
        handler: type,
        method: Callable[..., Any],
        name: str,
        description: str,
        authenticated: bool = True,
        needs_data: bool = False
    ) -> None:
        if str.startswith(path, '/'):
            raise ValueError(f"Path must not start with a slash: {path}")
        path = f'/a/{{pk}}/{path}'
        dependencies: list[fastapi.params.Depends] = [
            self._wrap_depends('subject', self.subject_class.resolve)
        ]

        self.add_api_route(
            methods=methods,
            endpoint=PluginEndpoint(
                'authorize',
                handler,
                method,
                authenticated=authenticated,
                needs_data=needs_data
            ),
            path=path,
            name=name,
            description=description,
            dependencies=dependencies,
            include_in_schema=False
        )

    def register_template_module(self, qualname: str) -> None:
        self.template_loaders.append(PackageLoader(qualname))

    async def authorize(
        self,
        response: Any = NotImplemented,
    ) -> fastapi.Response:
        """The **Authorization Endpoint** provides an interface for Resource Owners
        to interact with the Authorization Server in order to authenticate and
        authorize client requests.
        """
        assert response != NotImplemented
        return response

    async def deny(
        self,
        storage: Storage,
        request: PendingAuthorizationRequest,
        client: PendingAuthorizationClient
    ) -> fastapi.responses.RedirectResponse:
        """Deny an authorization request."""
        await storage.delete(request)
        return fastapi.responses.RedirectResponse(
            status_code=302,
            url=await request.deny(client)
        )

    async def keys(self) -> fastapi.responses.PlainTextResponse:
        await self.config.issuer.encryption_key
        await self.config.issuer.signing_key
        jwks = self.config.issuer.encryption_key | self.config.issuer.signing_key
        return fastapi.responses.PlainTextResponse(
            media_type='application/json;indent=2',
            headers={
                'Cache-Control': (
                    'max-age=43200, '
                    'stale-while-revalidate=43200, '
                    'stale-if-error=43200'
                )
            },
            content=json.dumps(
                jwks.model_dump(exclude_none=True),
                indent=2,
                sort_keys=True
            )
        )

    async def metadata(
        self,
        request: fastapi.Request
    ) -> fastapi.responses.PlainTextResponse:
        await asyncio.gather(
            asyncio.ensure_future(self.config.issuer.signing_key),
            asyncio.ensure_future(self.config.issuer.encryption_key),
        )
        enc_alg = self.config.issuer.encryption_key.get_encryption_algorithms()
        sig_alg = self.config.issuer.signing_key.get_signing_algorithms()
        cek_alg = [JWA.a128gcm, JWA.a192gcm, JWA.a256gcm]

        obj = ServerMetadata.model_validate({
            'issuer': self.config.issuer.id,
            'authorization_signing_alg_values_supported': sig_alg,
            'authorization_encryption_alg_values_supported': enc_alg,
            'authorization_encryption_enc_values_supported': cek_alg,
            'authorization_response_iss_parameter_supported': True,
            'jwks_uri': str(request.url_for('oauth2.jwks')),
            'id_token_signing_alg_values_supported': sig_alg,
            'id_token_encryption_alg_values_supported': enc_alg,
            'introspection_signing_alg_values_supported': sig_alg,
            'introspection_encryption_alg_values_supported': enc_alg,
            'nfv_token_signing_alg_values_supported': sig_alg,
            'nfv_token_encryption_alg_values_supported': enc_alg,
            'nfv_token_encryption_enc_values_supported': cek_alg,
            'request_object_signing_alg_values_supported': sig_alg,
            'request_object_encryption_alg_values_supported': enc_alg,
            'request_object_encryption_enc_values_supported': cek_alg,
            'token_endpoint_auth_signing_alg_values_supported': sig_alg,
            'userinfo_signing_alg_values_supported': sig_alg,
            'userinfo_encryption_alg_values_supported': enc_alg,
            'userinfo_encryption_enc_values_supported': cek_alg,
            'subject_types_supported': ['public', 'pairwise'],
            'token_endpoint': str(request.url_for('oauth2.token')),
            'userinfo_endpoint': str(request.url_for('oidc.userinfo'))
        })
        if self.config.has_authorize_endpoint():
            obj.authorization_endpoint = str(request.url_for('oauth2.authorize'))
        return fastapi.responses.PlainTextResponse(
            media_type='application/json',
            content=obj.model_dump_json(
                exclude_defaults=True,
                exclude_none=True,
                indent=2
            )
        )

    async def token(
        self,
        request: Request,
        handler: TokenEndpointHandler = NotImplemented,
    ) -> fastapi.Response:
        async with handler.transaction(request):
            return await handler.handle(request)

    async def userinfo(self, response: Any = NotImplemented) -> fastapi.Response:
        assert response != NotImplemented
        return response

    async def get_config(self) -> Config:
        return self.config

    async def get_keychain(self) -> CryptoKeyType:
        return self.signing_key

    async def get_signer(self) -> ISigner:
        if self.signer is None:
            self.signer = await self.config.ui.session_key # type: ignore
        return self.signer # type: ignore

    def _wrap_depends(
        self,
        name: str,
        dependency: Callable[..., Any]
    ) -> fastapi.params.Depends:
        """Set an attribute on the request object that has an instance of the
        given callable.
        """
        async def f(request: Request, dep: Any = fastapi.Depends(dependency)) -> None:
            setattr(request, name, dep)

        return fastapi.Depends(f)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import contextlib
import datetime
from typing import Any
from typing import Literal

import pydantic

from oauthx.lib.protocols import IStorage
from oauthx.lib.types import OIDCIssuerIdentifier
from oauthx.server.protocols import ISubject
from oauthx.server.protocols import ISubjectLogger
from oauthx.server.types import ClaimSetKey
from .claim import Claim


class Receipt(pydantic.BaseModel):
    client_id: str
    id: int
    obtained: datetime.datetime
    processed: set[str] = set()
    provider: OIDCIssuerIdentifier
    purpose: Literal['IDENTIFY', 'INVITE', 'LOGIN', 'VERIFY_ACCOUNT']
    received: set[str]
    request_id: int | None
    sub: int

    @property
    def claims(self) -> list[Claim]:
        return self._claims
    
    @property
    def claimset_id(self) -> ClaimSetKey:
        return ClaimSetKey(self.provider, self.sub)

    _claims: list[Claim] = pydantic.PrivateAttr(default_factory=list)
    _futures: list[asyncio.Future[Claim]] = pydantic.PrivateAttr(default_factory=list)
    _logger: ISubjectLogger = pydantic.PrivateAttr(default=None)
    _storage: IStorage | None = pydantic.PrivateAttr(default=None)
    _subject: ISubject | None = pydantic.PrivateAttr(default=None)

    @contextlib.asynccontextmanager
    async def transaction(self, subject: ISubject, storage: IStorage, logger: ISubjectLogger):
        if self._claims:
            raise ValueError("Transaction is already committed.")
        self._storage = storage
        self._subject = subject
        try:
            yield self
            logger.receipt(
                receipt_id=self.id,
                **self.model_dump(exclude={'id'})
            )
            self._claims = await asyncio.gather(*self._futures)
            await storage.persist(self)
        finally:
            self._futures = []
            self._storage = None
            self._subject = None

    def add(
        self,
        kind: Any,
        value: Any,
        *,
        issuer: Literal['self'] | OIDCIssuerIdentifier | None = None,
        ial: int = 0
    ) -> Claim | None:
        """Add a new claim to the :class:`Receipt`."""
        if value is None:
            return
        if kind not in self.received:
            raise ValueError(
                f"Claim '{kind}' is not registered as obtained with this "
                "receipt."
            )
        self.processed.add(kind)
        coro = self._create_claim(
            kind=kind,
            value=value,
            issuer=issuer,
            ial=ial
        )
        self._futures.append(asyncio.ensure_future(coro))

    async def _create_claim(self, kind: str, value: Any, **kwargs: Any) -> Claim:
        assert self._storage is not None
        assert self._subject is not None
        claim = Claim.model_validate({
            **kwargs,
            'kind': kind,
            'id': await self._storage.allocate_identifier(Claim),
            'obtained': self.obtained,
            'provider': self.provider,
            'receipt_id': self.id,
            'sub': self.sub,
            'value': value
        })
        await claim.mask(self._subject.get_masking_key())
        assert claim.is_masked()
        await claim.encrypt(self._subject)
        return claim


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

from oauthx.server.request import Request

from .base import BaseLoginRequestHandler


class NullLoginRequestHandler(BaseLoginRequestHandler):
    __module__: str = 'oauthx.server.handlers'

    async def handle(self, request: Request) -> fastapi.Response:
        return await request.render_to_response(
            template_names='oauthx/docs/implement-login.html.j2',
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import re
from typing import Any

import pydantic

from oauthx.lib.types import PKCEChallengeMethod
from oauthx.lib.types import RedirectURI
from oauthx.lib.types import ResponseType
from .baseauthorizationrequest import BaseAuthorizationRequest


class AuthorizationRequestParameters(BaseAuthorizationRequest):
    redirect_uri: RedirectURI | None
    response_type: ResponseType
    scope: set[str]
    state: str | None = None

    # RFC 7635 Proof Key for Code Exchange by OAuth Public Clients
    code_challenge: str | None = None
    code_challenge_method: PKCEChallengeMethod | None = None

    @pydantic.field_validator('scope', mode='before')
    def validate_scope(cls, scope: str | set[str] | None) -> set[str]:
        if isinstance(scope, str):
            scope = {str.strip(x) for x in re.split(r'\s+', scope)}
        scope = scope or set()
        return set(filter(bool, scope))

    @pydantic.model_validator(mode='before')
    def preprocess_parameters(cls, values: dict[str, Any]) -> dict[str, Any]:
        for param in {'request', 'request_uri'}:
            if not values.get(param):
                continue
            raise ValueError(f'The {param} parameter must not be provided.')
        return values


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi

from oauthx.lib.exceptions import Error
from oauthx.lib.exceptions import InvalidAuthorizationResponse
from oauthx.lib.exceptions import InvalidTokenResponse
from oauthx.lib.params import ClientAuthorizationState
from oauthx.lib.models import Error as ErrorModel
from oauthx.lib.models import ObtainedGrant as ObtainedGrantModel
from oauthx.lib.params import AuthorizationResponse
from oauthx.lib.params import Storage
from .provider import Provider


async def get(
    params: AuthorizationResponse,
    provider: Provider,
    state: ClientAuthorizationState,
    storage: Storage
) -> ObtainedGrantModel:
    if state is None:
        raise InvalidAuthorizationResponse(error='invalid_response')
    return_url = state.annotation('return-url')
    if return_url is None:
        await storage.delete(state)
        raise InvalidAuthorizationResponse(error='invalid_response')
    if params.is_error():
        assert isinstance(params.root, ErrorModel)
        await storage.delete(state)
        raise InvalidAuthorizationResponse(
            error=params.root.error,
            error_description=params.root.error_description,
            context={
                'authorize_url': state.get_authorize_url(),
                'provider': provider,
                'return_url': return_url,
            }
        )
    try:
        return await provider.authorization_code(params, state) # type: ignore
    except Error as exception:
        await storage.delete(state)
        raise InvalidTokenResponse(
            error=exception.error,
            error_description=exception.error_description,
            context={'provider': provider}
        )


AuthorizationCodeGrant: TypeAlias = Annotated[
    ObtainedGrantModel,
    fastapi.Depends(get)
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pathlib
import webbrowser
from typing import Iterable

import httpx

from oauthx.client.models import Client
from oauthx.lib.models import ObtainedGrant
from ..cli import LocalRedirection


class LocalAuthorization(httpx.Auth):
    client: Client
    grant: ObtainedGrant | None = None
    path: pathlib.Path | None = None

    def __init__(
        self,
        client: Client,
        scope: Iterable[str] | None = None,
        path: str | None = None
    ):
        self.client = client
        if path is not None:
            self.path = pathlib.Path(path).expanduser()
        self.scope = scope

        # Open the file at the given path to obtain the grant stored
        # locally.
        if self.path and self.path.exists():
            with open(self.path, 'r') as f:
                self.grant = ObtainedGrant.model_validate_json(f.read())

    async def async_auth_flow(self, request: httpx.Request):
        if not self.grant or self.grant.is_expired():
            port = self.client.get_local_port()
            state = await self.client.authorize(
                scope=self.scope,
                redirect_uri=f'http://127.0.0.1:{port}',
            )
            with LocalRedirection(self.client, state, port) as callback:
                webbrowser.open(state.authorize_url, new=2, autoraise=False)
            if callback.grant is None:
                callback.result.raise_for_status()
            assert callback.grant is not None
            self.grant = callback.grant
            if self.path:
                with open(self.path, 'w') as f:
                    f.write(self.grant.model_dump_json(indent=2))

        request.headers['Authorization'] = f'Bearer {self.grant.access_token}'
        yield request


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import enum


class ClientAuthenticationMethod(str, enum.Enum):
    none                        = 'none'
    client_secret_post          = 'client_secret_post'
    client_secret_basic         = 'client_secret_basic'
    client_secret_jwt           = 'client_secret_jwt'
    private_key_jwt             = 'private_key_jwt'
    tls_client_auth             = 'tls_client_auth'
    self_signed_tls_client_auth = 'self_signed_tls_client_auth'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .assertion import Assertion
from .authorizationcodegrant import AuthorizationCodeGrant
from .authorizationrequest import AuthorizationRequest
from .authorizationrequestobject import AuthorizationRequestObject
from .authorizationrequestparameters import AuthorizationRequestParameters
from .authorizationrequestreference import AuthorizationRequestReference
from .authorizationresponse import AuthorizationResponse
from .clientkey import ClientKey
from .clientauthorizationstate import ClientAuthorizationState
from .clientauthorizationstate import ClientAuthorizationStateKey
from .clientcredentialsgrant import ClientCredentialsGrant
from .error import Error
from .grant import Grant
from .jarmtoken import JARMToken
from .jwtbearerassertion import JWTBearerAssertion
from .jwtbearergrant import JWTBearerGrant
from .lazyservermetadata import LazyServerMetadata
from .obtainedgrant import ObtainedGrant
from .personalname import PersonalName
from .provider import Provider
from .redirectionparameters import RedirectionParameters
from .saml2bearerassertion import SAML2BearerAssertion
from .servermetadata import ServerMetadata


__all__: list[str] = [
    'Assertion',
    'AuthorizationCodeGrant',
    'AuthorizationRequest',
    'AuthorizationRequestObject',
    'AuthorizationRequestParameters',
    'AuthorizationRequestReference',
    'AuthorizationResponse',
    'ClientKey',
    'ClientAuthorizationState',
    'ClientAuthorizationStateKey',
    'ClientCredentialsGrant',
    'Error',
    'Grant',
    'JARMToken',
    'JWTBearerAssertion',
    'JWTBearerGrant',
    'LazyServerMetadata',
    'ObtainedGrant',
    'PersonalName',
    'Provider',
    'RedirectionParameters',
    'SAML2BearerAssertion',
    'ServerMetadata',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal
from typing import TypeAlias


SubjectClaimType: TypeAlias = Literal[
    'birthdate',
    'email',
    'family_name',
    'gender',
    'given_name',
    'middle_name',
    'name',
    'nickname',
    'phone_number',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi
from typing import Any

from oauthx.lib.types import RequestURI
from oauthx.lib.types import ResponseType

__all__: list[str] = [
    'CLIENT_ID',
    'REDIRECT_URI',
    'REQUEST',
    'REQUEST_URI',
    'RESPONSE_TYPE',
    'SCOPE',
    'STATE'
]

CLIENT_ID: str = fastapi.Query(
    default=...,
    title="Client ID",
    description="Identifies the client that is requesting an access token."
)

PROMPT: Any = fastapi.Query(
    default=None,
    title="Prompt",
    description=(
        "Space delimited, case sensitive list of ASCII string values that "
        "specifies whether the Authorization Server prompts the End-User "
        "for reauthentication and consent."
    )
)

REDIRECT_URI: str | None = fastapi.Query(
    default=None,
    title="Redirect URI",
    description=(
        "The client redirection endpoint. This URI **must** be priorly "
        "whitelisted for the client specified by `client_id`. If the "
        "client has multiple allowed redirect URIs and did not "
        "configure a default, then this parameter is **required**."
    )
)

REQUEST: str | None = fastapi.Query(
    default=None,
    alias='request',
    title="Request",
    description=(
        "A JSON Web Token (JWT) whose JWT Claims Set holds the "
        "JSON-encoded OAuth 2.0 authorization request parameters. "
        "Must not be used in combination with the `request_uri` "
        "parameter, and all other parameters except `client_id` "
        "must be absent.\n\n"
        "Confidential and credentialed clients must first sign "
        "the claims using their private key, and then encrypt the "
        "result with the public keys that are provided by the "
        "authorization server through the `jwks_uri` specified "
        "in its metadata."
    )
)

REQUEST_URI: RequestURI | None = fastapi.Query(
    default=None,
    title="Request URI",
    description=(
        "References a Pushed Authorization Request (PAR) or a remote "
        "object containing the authorization request.\n\n"
        "If the authorization request was pushed to this authorization "
        "server, then the format of the `request_uri` parameter is "
        "`urn:ietf:params:oauth:request_uri:<reference-value>`. "
        "Otherwise, it is an URI using the `https` scheme. If the "
        "`request_uri` parameter is a remote object, then the external "
        "domain must have been priorly whitelisted by the client."
    )
)

RESPONSE_MODE: str | None = fastapi.Query(
    default=None,
    title="Response mode",
    description=(
        "Informs the Authorization Server of the mechanism to be used "
        "for returning Authorization Response parameters from the "
        "Authorization Endpoint. This use of this parameter is "
        "**not recommended with a value that specifies the same "
        "Response Mode as the default Response Mode for the "
        "Response Type used."
    )
)

RESPONSE_TYPE: ResponseType | None = fastapi.Query(
    default=None,
    title="Response type",
    description="Specifies the response type.",
)

SCOPE: Any | None = fastapi.Query(
    default=None,
    title="Scope",
    description=(
        "The space-delimited scope that is requested by the client."
    )
)

STATE: str | None = fastapi.Query(
    default=None,
    alias='state',
    title="State",
    description=(
        "An opaque value used by the client to maintain state between the "
        "request and callback.  The authorization server includes this "
        "value when redirecting the user-agent back to the client."
    )
)